import tkinter as tk
from tkinter import ttk

from thonny import get_workbench, languages
from thonny.config_ui import ConfigurationPage


class GeneralConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)

        self.add_checkbox(
            "general.single_instance", _("Allow only single Thonny instance"), row=1, columnspan=2
        )
        self.add_checkbox(
            "general.event_logging", _("Log program usage events"), row=4, columnspan=2
        )
        self.add_checkbox(
            "file.reopen_all_files",
            _("Reopen all files from previous session"),
            row=5,
            columnspan=2,
        )
        self.add_checkbox(
            "general.disable_notification_sound",
            _("Disable notification sound"),
            row=6,
            columnspan=2,
        )
        self.add_checkbox(
            "general.debug_mode",
            _("Debug mode (provides more detailed diagnostic logs)"),
            row=7,
            columnspan=2,
        )

        # language
        self._language_name_var = tk.StringVar(
            value=languages.LANGUAGES_DICT.get(get_workbench().get_option("general.language"), "")
        )

        self._language_label = ttk.Label(self, text=_("Language"))
        self._language_label.grid(row=10, column=0, sticky=tk.W, padx=(0, 10), pady=(10, 0))
        self._language_combo = ttk.Combobox(
            self,
            width=20,
            exportselection=False,
            textvariable=self._language_name_var,
            state="readonly",
            height=15,
            values=list(languages.LANGUAGES_DICT.values()),
        )
        self._language_combo.grid(row=10, column=1, sticky=tk.W, pady=(10, 0))

        # Mode
        ttk.Label(self, text=_("UI mode")).grid(
            row=20, column=0, sticky=tk.W, padx=(0, 10), pady=(10, 0)
        )
        self.add_combobox(
            "general.ui_mode",
            ["simple", "regular", "expert"],
            row=20,
            column=1,
            pady=(10, 0),
            width=10,
        )

        # scaling
        self._scaling_label = ttk.Label(self, text=_("UI scaling factor"))
        self._scaling_label.grid(row=30, column=0, sticky=tk.W, padx=(0, 10), pady=(10, 0))
        scalings = sorted({0.5, 0.75, 1.0, 1.25, 1.33, 1.5, 2.0, 2.5, 3.0, 4.0})
        self.add_combobox("general.scaling", scalings, row=30, column=1, pady=(10, 0), width=10)

        self._font_scaling_var = get_workbench().get_variable("general.font_scaling_mode")
        self._font_scaling_label = ttk.Label(self, text=_("Font scaling mode"))
        self._font_scaling_label.grid(row=40, column=0, sticky=tk.W, padx=(0, 10), pady=(10, 0))
        self.add_combobox(
            "general.font_scaling_mode",
            ["default", "extra", "automatic"],
            row=40,
            column=1,
            pady=(10, 0),
            width=10,
        )

        reopen_label = ttk.Label(
            self,
            text=_("NB! Restart Thonny after changing these options!"),
            font="BoldTkDefaultFont",
        )
        reopen_label.grid(row=100, column=0, sticky=tk.W, pady=20, columnspan=2)

        self.columnconfigure(1, weight=1)

    def apply(self):
        get_workbench().set_option(
            "general.language", languages.get_language_code_by_name(self._language_name_var.get())
        )


def load_plugin() -> None:
    get_workbench().add_configuration_page("general", _("General"), GeneralConfigurationPage, 10)



class Pin:
    ""


class Processor:
    ""
    frequency = None
    temperature = None
    uid = None


class RunMode:
    ""
    BOOTLOADER = None
    NORMAL = None
    SAFE_MODE = None


cpu = None


def delay_us():
    pass


def disable_interrupts():
    pass


def enable_interrupts():
    pass


nvm = None


def on_next_reset():
    pass


pin = None


def reset():
    pass



import queue
from thonny.misc_utils import TimeHelper
from queue import Queue
import re


class MicroPythonConnection:
    """Utility class for using Serial or WebSocket connection
    
    Uses background thread to read from the source as soon as possible
    to avoid loss of data (because buffer overflow or the device discarding 
    unread data).

    Allows writing with delays after each n bytes.
    
    Allows unreading data.    
    """

    def __init__(self):
        self._read_queue = Queue()  # populated by reader thread
        self._read_buffer = bytearray()  # used for unreading and postponing bytes
        self.num_bytes_received = 0
        self._error = None

    def read(self, size, timeout=1):
        if timeout == 0:
            raise TimeoutError()

        timer = TimeHelper(timeout)

        while len(self._read_buffer) < size:
            self._check_for_error()

            try:
                self._read_buffer.extend(self._read_queue.get(True, timer.time_left))
            except queue.Empty:
                raise TimeoutError("Reaction timeout. Bytes read: %s" % self._read_buffer)

        try:
            data = self._read_buffer[:size]
            return data
        finally:
            del self._read_buffer[:size]

    def soft_read_until(self, terminator, timeout=1000000):
        return self.read_until(terminator, timeout, timeout_is_soft=True)

    def read_until(self, terminator, timeout=1000000, timeout_is_soft=False):
        timer = TimeHelper(timeout)

        if isinstance(terminator, str):
            terminator = re.compile(re.escape(terminator))

        match = None
        while True:
            self._check_for_error()

            match = re.search(terminator, self._read_buffer)
            if match:
                break

            try:
                data = self._read_queue.get(True, timer.time_left)
                # print("RR", repr(data), file=sys.stderr)
                assert len(data) > 0
                self._read_buffer.extend(data)
            except queue.Empty:
                if timeout_is_soft:
                    break
                else:
                    raise TimeoutError("Reaction timeout. Bytes read: %s" % self._read_buffer)

        if match:
            size = match.end()
        else:
            assert timeout_is_soft
            size = len(self._read_buffer)

        data = self._read_buffer[:size]
        del self._read_buffer[:size]
        return data

    def _fetch_to_buffer(self):
        while not self._read_queue.empty():
            self._read_buffer.extend(self._read_queue.get(True))

    def peek_incoming(self):
        self._fetch_to_buffer()
        return self._read_buffer

    def read_all(self):
        self._fetch_to_buffer()

        if len(self._read_buffer) == 0:
            self._check_for_error()

        try:
            return self._read_buffer
        finally:
            self._read_buffer = bytearray()

    def _check_for_error(self):
        if self._error is None:
            return

        raise ConnectionClosedException(self._error)

    def unread(self, data):
        self._read_buffer = data + self._read_buffer

    def write(self, data, block_size=32, delay=0.01):
        raise NotImplementedError()

    def _log_data(self, data):
        print(
            data.decode("Latin-1")
            .replace("\r\n", "\n")
            .replace("\x01", "①")
            .replace("\x02", "②")
            .replace("\x03", "③")
            .replace("\x04", "④"),
            end="",
        )

    def incoming_is_empty(self):
        return self._read_queue.empty() and len(self._read_buffer) == 0

    def outgoing_is_empty(self):
        return True

    def buffers_are_empty(self):
        return self.incoming_is_empty() and self.outgoing_is_empty()

    def reset_input_buffer(self):
        return self.read_all()

    def reset_output_buffer(self):
        pass

    def close(self):
        raise NotImplementedError()


class ConnectionFailedException(Exception):
    pass


class ConnectionClosedException(Exception):
    pass



import os.path
from thonny.tktextext import TextFrame, EnhancedText
from thonny import get_workbench, ui_utils, THONNY_USER_DIR
from thonny.ui_utils import TextMenu


class NotesText(EnhancedText):
    def __init__(self, master, **kw):
        EnhancedText.__init__(self, master=master, wrap="word", undo=True, **kw)
        self.context_menu = TextMenu(self)

    def on_secondary_click(self, event=None):
        super().on_secondary_click(event=event)
        self.context_menu.tk_popup(event.x_root, event.y_root)


class NotesView(TextFrame):
    def __init__(self, master):
        self.filename = os.path.join(THONNY_USER_DIR, "user_notes.txt")
        super().__init__(
            master, text_class=NotesText, horizontal_scrollbar_class=ui_utils.AutoScrollbar
        )

        self.load_content()
        self.text.edit_reset()

        get_workbench().bind("ToplevelResponse", self.save_content, True)

    def load_content(self):
        if not os.path.isfile(self.filename):
            self.text.insert(
                "1.0",
                _(
                    "This box is meant for your working notes -- assignment instructions, "
                    + "code snippets, whatever.\n\n"
                    + "Everything will be saved automatically "
                    + "and loaded when you open Thonny next time.\n\n"
                    + "Feel free to delete this text to make room for your own notes."
                ),
            )
            return

        with open(self.filename, encoding="utf-8") as fp:
            content = fp.read()

        self.text.delete("1.0", "end")
        self.text.insert("1.0", content)
        self.text.mark_set("insert", "1.0")
        self.text.see("1.0")

    def save_content(self, event=None):
        with open(self.filename, "w", encoding="utf-8") as fp:
            fp.write(self.text.get("1.0", "end-1c"))

    def destroy(self):
        self.save_content()
        super().destroy()


def load_plugin():
    get_workbench().add_view(NotesView, _("Notes"), "ne", default_position_key="zz")



BASE_LANGUAGE_CODE = "en_US"
BASE_LANGUAGE_NAME = "English"

LANGUAGES_DICT = {
    "de_DE": "Deutsch [ALPHA]",
    "et_EE": "Eesti",
    BASE_LANGUAGE_CODE: BASE_LANGUAGE_NAME,
    "es_ES": "Español [ALPHA]",
    "fr_FR": "Français [ALPHA]",
    "nl_NL": "Nederlands [BETA]",
    "pt_PT": "Português (PT) [BETA]",
    "pt_BR": "Português (BR) [ALPHA]",
    "ru_RU": "Русский [BETA]",
    "uk_UA": "Українська [BETA]",
    "zh_TW": "繁體中文-TW [BETA]",
    "zh_CN": "简体中文  [BETA]",
    "ja_JP": "日本語  [ALPHA]",
    "lt_LT": "Lietuvių [BETA]",
}


def get_language_code_by_name(name):
    for code in LANGUAGES_DICT:
        if LANGUAGES_DICT[code] == name:
            return code



import os.path
from threading import Thread
from tkinter import messagebox

from thonny import get_workbench, get_runner
import logging

_server_started = False


def _start_debug_enabled():
    return (
        get_workbench().get_editor_notebook().get_current_editor() is not None
        and "debug" in get_runner().get_supported_features()
    )


def start_server():
    try:
        from birdseye import server

        server.app.run(
            port=get_workbench().get_option("run.birdseye_port"), debug=False, use_reloader=False
        )
    except Exception:
        logging.getLogger("thonny").exception("Problem running Birdseye server")


def debug_with_birdseye():
    global _server_started

    try:
        import birdseye  # @UnusedImport
    except ImportError:
        if messagebox.askyesno(
            _("About Birdseye"),
            _(
                "Birdseye is a Python debugger which needs to be installed separately.\n\n"
                + "Do you want to open the help page and learn more?"
            ),
        ):
            get_workbench().open_help_topic("birdseye")

        return

    if not _server_started:
        _server_started = True
        Thread(target=start_server, daemon=True).start()

    os.environ["BIRDSEYE_PORT"] = str(get_workbench().get_option("run.birdseye_port"))
    get_runner().execute_current("Birdseye")


# order_key makes the plugin to be loaded later than other same tier plugins
# This way it gets positioned after main debug commands in the Run menu
load_order_key = "zz"


def load_plugin():
    get_workbench().set_default("run.birdseye_port", 7777)
    get_workbench().add_command(
        "birdseye",
        "run",
        _("Debug current script (birdseye)"),
        debug_with_birdseye,
        caption="birdseye",
        tester=_start_debug_enabled,
        default_sequence="<Control-B>",
        group=10,
        image=os.path.join(os.path.dirname(__file__), "..", "res", "birdseye.png"),
    )



import textwrap
import tkinter as tk
from tkinter import font as tk_font
from tkinter import ttk

from thonny import get_workbench, tktextext, ui_utils
from thonny.codeview import CodeView
from thonny.config_ui import ConfigurationPage
from thonny.ui_utils import create_string_var, scrollbar_style
from thonny.shell import BaseShellText


class ThemeAndFontConfigurationPage(ConfigurationPage):
    def __init__(self, master):

        ConfigurationPage.__init__(self, master)

        self._init_themes()
        self._init_fonts()
        self._init_previews()

        self.columnconfigure(2, weight=1)
        self.columnconfigure(4, weight=1)

        self.rowconfigure(31, weight=1)
        self.rowconfigure(21, weight=1)

    def _init_themes(self):
        self._original_ui_theme = get_workbench().get_option("view.ui_theme")
        self._original_syntax_theme = get_workbench().get_option("view.syntax_theme")

        self._ui_theme_variable = create_string_var(
            self._original_ui_theme, modification_listener=self._update_appearance
        )
        self._syntax_theme_variable = create_string_var(
            self._original_syntax_theme, modification_listener=self._update_appearance
        )

        ttk.Label(self, text=_("UI theme")).grid(
            row=1, column=1, sticky="w", pady=(0, 10), padx=(0, 5)
        )
        self._ui_theme_combo = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._ui_theme_variable,
            state="readonly",
            height=15,
            values=get_workbench().get_usable_ui_theme_names(),
        )
        self._ui_theme_combo.grid(row=1, column=2, sticky="nwe", pady=(0, 5))

        ttk.Label(self, text=_("Syntax theme")).grid(
            row=2, column=1, sticky="w", pady=(0, 10), padx=(0, 5)
        )
        self._syntax_theme_combo = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._syntax_theme_variable,
            state="readonly",
            height=15,
            values=get_workbench().get_syntax_theme_names(),
        )
        self._syntax_theme_combo.grid(row=2, column=2, sticky="nwe", pady=(0, 5))

    def _init_fonts(self):
        self._original_editor_family = get_workbench().get_option("view.editor_font_family")
        self._original_editor_size = get_workbench().get_option("view.editor_font_size")
        self._original_io_family = get_workbench().get_option("view.io_font_family")
        self._original_io_size = get_workbench().get_option("view.io_font_size")

        self._editor_family_variable = create_string_var(
            self._original_editor_family, modification_listener=self._update_appearance
        )
        self._editor_size_variable = create_string_var(
            self._original_editor_size, modification_listener=self._update_appearance
        )
        self._io_family_variable = create_string_var(
            self._original_io_family, modification_listener=self._update_appearance
        )
        self._io_size_variable = create_string_var(
            self._original_io_size, modification_listener=self._update_appearance
        )

        ttk.Label(self, text=_("Editor font")).grid(
            row=1, column=3, sticky="w", pady=(0, 5), padx=(25, 5)
        )
        editor_family_combo = ttk.Combobox(
            self,
            exportselection=False,
            state="readonly",
            height=15,
            textvariable=self._editor_family_variable,
            values=self._get_families_to_show(),
        )
        editor_family_combo.grid(row=1, column=4, sticky="nwe", pady=(0, 5))
        editor_size_combo = ttk.Combobox(
            self,
            width=4,
            exportselection=False,
            textvariable=self._editor_size_variable,
            state="readonly",
            height=15,
            values=[str(x) for x in range(3, 73)],
        )
        editor_size_combo.grid(row=1, column=5, sticky="nwe", pady=(0, 5), padx=(5, 0))

        ttk.Label(self, text=_("IO font")).grid(
            row=2, column=3, sticky="w", pady=(0, 5), padx=(25, 5)
        )
        io_family_combo = ttk.Combobox(
            self,
            exportselection=False,
            state="readonly",
            height=15,
            textvariable=self._io_family_variable,
            values=self._get_families_to_show(),
        )
        io_family_combo.grid(row=2, column=4, sticky="nwe", pady=(0, 5))

        io_size_combo = ttk.Combobox(
            self,
            width=4,
            exportselection=False,
            textvariable=self._io_size_variable,
            state="readonly",
            height=15,
            values=[str(x) for x in range(3, 73)],
        )
        io_size_combo.grid(row=2, column=5, sticky="nwe", pady=(0, 5), padx=(5, 0))

    def _init_previews(self):
        ttk.Label(self, text=_("Preview")).grid(
            row=20, column=1, sticky="w", pady=(5, 2), columnspan=5
        )
        self._preview_codeview = CodeView(
            self, height=6, font="EditorFont", relief="groove", borderwidth=1
        )

        self._preview_codeview.set_content(
            textwrap.dedent(
                """
            def foo(bar):
                if bar is None: # """
                + _("This is a comment")
                + """
                    print('"""
                + _("The answer is")
                + """', 33)

            """
                + _("unclosed_string")
                + ''' = "'''
                + _("blah, blah")
                + "\n"
            ).strip()
        )
        self._preview_codeview.grid(row=21, column=1, columnspan=5, sticky=tk.NSEW)

        self._shell_preview = tktextext.TextFrame(
            self,
            text_class=BaseShellText,
            height=4,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            relief="groove",
            borderwidth=1,
            font="EditorFont",
        )
        self._shell_preview.grid(row=31, column=1, columnspan=5, sticky=tk.NSEW, pady=(5, 5))
        self._shell_preview.text.set_read_only(True)
        self._insert_shell_text()

        ttk.Label(
            self,
            text=_("NB! Some style elements change only after restarting Thonny!"),
            font="BoldTkDefaultFont",
        ).grid(row=40, column=1, columnspan=5, sticky="w", pady=(5, 0))

    def apply(self):
        # don't do anything, as preview already did the thing
        return

    def cancel(self):
        if (
            getattr(self._editor_family_variable, "modified")
            or getattr(self._editor_size_variable, "modified")
            or getattr(self._ui_theme_variable, "modified")
            or getattr(self._syntax_theme_variable, "modified")
        ):
            get_workbench().set_option("view.ui_theme", self._original_ui_theme)
            get_workbench().set_option("view.syntax_theme", self._original_syntax_theme)
            get_workbench().set_option("view.editor_font_size", self._original_editor_size)
            get_workbench().set_option("view.editor_font_family", self._original_editor_family)
            get_workbench().set_option("view.io_font_size", self._original_io_size)
            get_workbench().set_option("view.io_font_family", self._original_io_family)
            get_workbench().reload_themes()
            get_workbench().update_fonts()

    def _update_appearance(self):
        get_workbench().set_option("view.ui_theme", self._ui_theme_variable.get())
        get_workbench().set_option("view.syntax_theme", self._syntax_theme_variable.get())
        get_workbench().set_option("view.editor_font_size", int(self._editor_size_variable.get()))
        get_workbench().set_option("view.editor_font_family", self._editor_family_variable.get())
        get_workbench().set_option("view.io_font_size", int(self._io_size_variable.get()))
        get_workbench().set_option("view.io_font_family", self._io_family_variable.get())
        get_workbench().reload_themes()
        get_workbench().update_fonts()

    def _insert_shell_text(self):
        text = self._shell_preview.text
        text._insert_prompt()
        text.direct_insert("end", "%Run demo.py\n", ("magic", "before_io"))
        text.tag_add("before_io", "1.0", "1.0 lineend")
        text.direct_insert("end", _("Enter an integer") + ": ", ("io", "stdout"))
        text.direct_insert("end", "2.5\n", ("io", "stdin"))
        text.direct_insert(
            "end", "ValueError: invalid literal for int() with base 10: '2.5'\n", ("io", "stderr")
        )

    def _get_families_to_show(self):
        # In Linux, families may contain duplicates (actually different fonts get same names)
        return sorted(set(filter(lambda name: name[0].isalpha(), tk_font.families())))


def load_plugin() -> None:
    get_workbench().add_configuration_page(
        "theme", _("Theme & Font"), ThemeAndFontConfigurationPage, 40
    )



RATE_1MBIT = 0
RATE_250KBIT = 2
RATE_2MBIT = 1


def config():
    pass


def off():
    pass


def on():
    pass


def receive():
    pass


def receive_bytes():
    pass


def receive_bytes_into():
    pass


def receive_full():
    pass


def reset():
    pass


def send():
    pass


def send_bytes():
    pass



from thonny.assistance import ProgramAnalyzer, add_program_analyzer
from thonny import get_runner, rst_utils
from thonny.running import CPythonProxy
import logging
import os
from thonny.common import is_same_path


known_stdlib_modules = {
    # Compiled from https://docs.python.org/3.7/py-modindex.html
    "__future__",
    "__main__",
    "_dummy_thread",
    "_thread",
    "abc",
    "aifc",
    "argparse",
    "array",
    "ast",
    "asynchat",
    "asyncio",
    "asyncore",
    "atexit",
    "audioop",
    "base64",
    "bdb",
    "binascii",
    "binhex",
    "bisect",
    "builtins",
    "bz2",
    "calendar",
    "cgi",
    "cgitb",
    "chunk",
    "cmath",
    "cmd",
    "code",
    "codecs",
    "codeop",
    "collections",
    "colorsys",
    "compileall",
    "concurrent",
    "configparser",
    "contextlib",
    "contextvars",
    "copy",
    "copyreg",
    "cProfile",
    "crypt",
    "csv",
    "ctypes",
    "curses",
    "dataclasses",
    "datetime",
    "dbm",
    "decimal",
    "difflib",
    "dis",
    "distutils",
    "doctest",
    "dummy_threading",
    "email",
    "encodings",
    "ensurepip",
    "enum",
    "errno",
    "faulthandler",
    "fcntl",
    "filecmp",
    "fileinput",
    "fnmatch",
    "formatter",
    "fractions",
    "ftplib",
    "functools",
    "gc",
    "getopt",
    "getpass",
    "gettext",
    "glob",
    "grp",
    "gzip",
    "hashlib",
    "heapq",
    "hmac",
    "html",
    "http",
    "imaplib",
    "imghdr",
    "imp",
    "importlib",
    "inspect",
    "io",
    "ipaddress",
    "itertools",
    "json",
    "keyword",
    "lib2to3",
    "linecache",
    "locale",
    "logging",
    "lzma",
    "macpath",
    "mailbox",
    "mailcap",
    "marshal",
    "math",
    "mimetypes",
    "mmap",
    "modulefinder",
    "msilib",
    "msvcrt",
    "multiprocessing",
    "netrc",
    "nis",
    "nntplib",
    "numbers",
    "operator",
    "optparse",
    "os",
    "ossaudiodev",
    "parser",
    "pathlib",
    "pdb",
    "pickle",
    "pickletools",
    "pipes",
    "pkgutil",
    "platform",
    "plistlib",
    "poplib",
    "posix",
    "pprint",
    "profile",
    "pstats",
    "pty",
    "pwd",
    "py_compile",
    "pyclbr",
    "pydoc",
    "queue",
    "quopri",
    "random",
    "re",
    "readline",
    "reprlib",
    "resource",
    "rlcompleter",
    "runpy",
    "sched",
    "secrets",
    "select",
    "selectors",
    "shelve",
    "shlex",
    "shutil",
    "signal",
    "site",
    "smtpd",
    "smtplib",
    "sndhdr",
    "socket",
    "socketserver",
    "spwd",
    "sqlite3",
    "ssl",
    "stat",
    "statistics",
    "string",
    "stringprep",
    "struct",
    "subprocess",
    "sunau",
    "symbol",
    "symtable",
    "sys",
    "sysconfig",
    "syslog",
    "tabnanny",
    "tarfile",
    "telnetlib",
    "tempfile",
    "termios",
    "test",
    "textwrap",
    "threading",
    "time",
    "timeit",
    "tkinter",
    "token",
    "tokenize",
    "trace",
    "traceback",
    "tracemalloc",
    "tty",
    "turtle",
    "turtledemo",
    "types",
    "typing",
    "unicodedata",
    "unittest",
    "urllib",
    "uu",
    "uuid",
    "venv",
    "warnings",
    "wave",
    "weakref",
    "webbrowser",
    "winreg",
    "winsound",
    "wsgiref",
    "xdrlib",
    "xml",
    "xmlrpc",
    "zipapp",
    "zipfile",
    "zipimport",
    "zlib",
}


class ProgramNamingAnalyzer(ProgramAnalyzer):
    def start_analysis(self, main_file_path, imported_file_paths):
        self.completion_handler(self, list(self._get_warnings(main_file_path)))

    def _get_warnings(self, main_file_path):
        # TODO: current dir may be different
        main_file_dir = os.path.dirname(main_file_path)
        if not os.path.isdir(main_file_dir):
            return []

        library_modules = known_stdlib_modules | self._get_3rd_party_modules()

        for item in os.listdir(main_file_dir):
            full_path = os.path.join(main_file_dir, item)
            if item.endswith(".py") and item[:-3] in library_modules:

                if is_same_path(full_path, main_file_path):
                    prelude = "Your program file is named '%s'." % item
                    rename_hint = " (*File → Rename…* )"
                else:
                    prelude = (
                        "Your working directory `%s <%s>`__ contains a file named '%s'.\n\n"
                        % (rst_utils.escape(main_file_dir), rst_utils.escape(main_file_dir), item)
                    )
                    rename_hint = ""

                yield {
                    "filename": full_path,
                    "lineno": 0,
                    "symbol": "file-shadows-library-module",
                    "msg": "Possibly bad file name",
                    "explanation_rst": prelude
                    + "\n\n"
                    + "When you try to import library module ``%s``, your file will be imported instead.\n\n"
                    % item[:-3]
                    + "Rename your '%s'%s to make the library module visible again."
                    % (item, rename_hint),
                    "group": "warnings",
                    "relevance": 5,
                }

    def _get_3rd_party_modules(self):
        proxy = get_runner().get_backend_proxy()
        if not isinstance(proxy, CPythonProxy):
            return []

        try:
            sys_path = proxy.get_sys_path()
        except Exception:
            logging.exception("Can't get sys path from proxy")
            return []

        module_names = set()
        for item in sys_path:
            if os.path.isdir(item) and ("site-packages" in item or "dist-packages" in item):
                module_names.update(self._get_module_names(item))
                for name in os.listdir(item):
                    if "-" not in name:
                        module_names.add(name.replace(".py", ""))

        return module_names

    def _get_module_names(self, dir_path):
        result = set()
        for name in os.listdir(dir_path):
            if "-" not in name:
                result.add(name.replace(".py", ""))
        return result


def load_plugin():
    add_program_analyzer(ProgramNamingAnalyzer)



"""Utils to handle different jedi versions

Because of Debian Stretch, Thonny needs to support jedi 0.10,
which doesn't use separate parso
"""


def import_python_tree():
    # pylint: disable=import-error,no-name-in-module

    if get_version_tuple() < (0, 11, 0):
        # make sure not to import parso in this case, even if it exits
        from jedi.parser.python import tree  # @UnresolvedImport @UnusedImport
    else:
        # assume older versions, which use parso
        from parso.python import tree  # @UnresolvedImport @UnusedImport @Reimport

    return tree


def get_params(func_node):
    if hasattr(func_node, "get_params"):
        # parso
        return func_node.get_params()
    else:
        # older jedi
        return func_node.params


def get_parent_scope(node):
    try:
        # jedi 0.11
        from jedi import parser_utils

        return parser_utils.get_parent_scope(node)
    except ImportError:
        # Older versions
        return node.get_parent_scope()


def get_statement_of_position(node, pos):
    try:
        # jedi 0.11
        # pylint: disable=redefined-outer-name
        from jedi.parser_utils import get_statement_of_position

        return get_statement_of_position(node, pos)
    except ImportError:
        # Older versions
        return node.get_statement_for_position(pos)


def get_module_node(script):
    if hasattr(script, "_module_node"):  # Jedi 0.12
        return script._module_node
    elif hasattr(script, "_get_module_node"):  # Jedi 0.10 - 0.11
        return script._get_module_node()
    elif hasattr(script, "_get_module"):
        return script._get_module()
    else:
        return script._parser.module()


def is_scope(node):
    try:
        # jedi 0.11 and older
        from jedi import parser_utils

        return parser_utils.is_scope(node)
    except ImportError:
        # Older versions
        return node.is_scope()


def get_name_of_position(obj, position):
    if hasattr(obj, "get_name_of_position"):
        # parso
        return obj.get_name_of_position(position)
    else:
        # older jedi
        return obj.name_for_position(position)


def parse_source(source):
    try:
        import parso

        return parso.parse(source)
    except ImportError:
        import jedi

        script = jedi.Script(source)
        return get_module_node(script)


def get_version_tuple():
    import jedi

    nums = []
    for item in jedi.__version__.split("."):
        try:
            nums.append(int(item))
        except Exception:
            nums.append(0)

    return tuple(nums)



LANDSCAPE = 1
LANDSCAPE_UPSIDEDOWN = 3


class LCD160CR:
    ""

    def _fcmd2():
        pass

    def _fcmd2b():
        pass

    def _send():
        pass

    def _waitfor():
        pass

    def clip_line():
        pass

    def dot():
        pass

    def dot_no_clip():
        pass

    def erase():
        pass

    def fast_spi():
        pass

    def feed_wdt():
        pass

    def get_line():
        pass

    def get_pixel():
        pass

    def get_touch():
        pass

    def iflush():
        pass

    def is_touched():
        pass

    def jpeg():
        pass

    def jpeg_data():
        pass

    def jpeg_start():
        pass

    def line():
        pass

    def line_no_clip():
        pass

    def oflush():
        pass

    def poly_dot():
        pass

    def poly_line():
        pass

    def rect():
        pass

    def rect_interior():
        pass

    def rect_interior_no_clip():
        pass

    def rect_no_clip():
        pass

    def rect_outline():
        pass

    def rect_outline_no_clip():
        pass

    def reset():
        pass

    def rgb():
        pass

    def save_to_flash():
        pass

    def screen_dump():
        pass

    def screen_load():
        pass

    def set_brightness():
        pass

    def set_font():
        pass

    def set_i2c_addr():
        pass

    def set_orient():
        pass

    def set_pen():
        pass

    def set_pixel():
        pass

    def set_pos():
        pass

    def set_power():
        pass

    def set_scroll():
        pass

    def set_scroll_buf():
        pass

    def set_scroll_win():
        pass

    def set_scroll_win_param():
        pass

    def set_spi_win():
        pass

    def set_startup_deco():
        pass

    def set_text_color():
        pass

    def set_uart_baudrate():
        pass

    def show_framebuf():
        pass

    def touch_config():
        pass

    def write():
        pass


PORTRAIT = 0
PORTRAIT_UPSIDEDOWN = 2
STARTUP_DECO_INFO = 2
STARTUP_DECO_MLOGO = 1
STARTUP_DECO_NONE = 0
_uart_baud_table = None


def calcsize():
    pass


def const():
    pass


machine = None


def pack_into():
    pass


def sleep_ms():
    pass


uerrno = None



import sys
import os.path
from thonny.plugins.micropython import (
    MicroPythonProxy,
    MicroPythonConfigPage,
    add_micropython_backend,
)
from thonny import get_workbench, ui_utils
from thonny.ui_utils import FileCopyDialog, CommonDialog, ems_to_pixels
from thonny.misc_utils import find_volume_by_name
import shutil
from time import sleep
from tkinter import ttk, messagebox
import threading
from urllib.request import urlopen
import json

LATEST_RELEASE_URL = "https://api.github.com/repos/bbcmicrobit/micropython/releases/latest"


class MicrobitProxy(MicroPythonProxy):
    def _start_background_process(self, clean=None):
        # NB! Sometimes disconnecting and reconnecting (on macOS?)
        # too quickly causes anomalies
        # https://github.com/pyserial/pyserial/issues/176
        # In my Sierra, Calliope and micro:bit seemed to soft-reboot
        # when reconnected too quickly.

        if clean and sys.platform == "darwin":
            sleep(1.0)

        super()._start_background_process(clean)

    def supports_directories(self):
        return False

    @property
    def consider_unknown_devices(self):
        return False

    @property
    def known_usb_vids_pids(self):
        """Copied from https://github.com/mu-editor/mu/blob/master/mu/modes/adafruit.py"""
        return {(0x0D28, 0x0204)}  # Adafruit CRICKit M0


class MicrobitConfigPage(MicroPythonConfigPage):
    def _get_usb_driver_url(self):
        return "https://microbit-micropython.readthedocs.io/en/latest/devguide/repl.html"

    def _get_flashing_frame(self):
        frame = super()._get_flashing_frame()
        self._flashing_label.configure(
            text=_("Make sure MicroPython has been installed to your micro:bit.")
            + "\n("
            + _("Don't forget that main.py only works without embedded main script.")
            + ")"
        )
        return frame

    def _has_flashing_dialog(self):
        return True

    def _open_flashing_dialog(self):
        mount_path = find_volume_by_name(
            "MICROBIT",
            not_found_msg=_("Could not find disk '%s'.")
            + "\n"
            + _("Make sure you have micro:bit plugged in!")
            + "\n\n"
            + _("Do you want to continue and locate the disk yourself?"),
        )
        """
        mount_path = "C:\\"
        """
        if mount_path is None:
            return

        dlg = FlashingDialog(get_workbench(), mount_path)
        ui_utils.show_dialog(dlg)


class FlashingDialog(CommonDialog):
    def __init__(self, master, target_dir):
        self._release_info = None
        self._hex_url = None
        self._hex_size = None
        self._target_dir = target_dir
        self._bytes_copied = 0

        self._state = "starting"

        super().__init__(master)

        main_frame = ttk.Frame(self)  # To get styled background
        main_frame.grid(row=0, column=0, sticky="nsew")
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        main_frame.columnconfigure(1, weight=1)

        self.title(_("Install latest MicroPython to BBC micro:bit"))

        target_caption_label = ttk.Label(main_frame, text=_("micro:bit location:"))
        target_caption_label.grid(row=0, column=0, padx=15, pady=(15, 0), sticky="w")
        target_label = ttk.Label(main_frame, text=self._target_dir)
        target_label.grid(row=0, column=1, padx=15, pady=(15, 0), sticky="w", columnspan=2)

        version_caption_label = ttk.Label(main_frame, text=_("Version to be installed:"))
        version_caption_label.grid(row=1, column=0, sticky="w", padx=15, pady=(0, 15))
        self._version_label = ttk.Label(main_frame, text=_("please wait") + " ...")
        self._version_label.grid(row=1, column=1, columnspan=2, padx=15, pady=(0, 15), sticky="w")

        self._state_label = ttk.Label(
            main_frame, text=_("NB! All files on micro:bit will be deleted!")
        )
        self._state_label.grid(row=2, column=0, columnspan=3, sticky="w", padx=15, pady=(0, 15))

        self._progress_bar = ttk.Progressbar(main_frame, length=ems_to_pixels(30))
        self._progress_bar.grid(row=3, column=0, columnspan=3, sticky="nsew", padx=15, pady=0)

        self._install_button = ttk.Button(
            main_frame, text=_("Install"), command=self._start_installing
        )
        self._install_button.grid(row=4, column=0, columnspan=2, sticky="ne", padx=0, pady=15)

        self._close_button = ttk.Button(main_frame, text=_("Cancel"), command=self._close)
        self._close_button.grid(row=4, column=2, sticky="ne", padx=15, pady=15)
        self._progress_bar.focus_set()

        main_frame.columnconfigure(1, weight=1)

        self.bind("<Escape>", self._close, True)  # escape-close only if process has completed
        self.protocol("WM_DELETE_WINDOW", self._close)

        self._start_downloading_release_info()
        self._update_state()

    def _locate_microbit(self):
        pass

    def _start_downloading_release_info(self):
        def work():
            with urlopen(LATEST_RELEASE_URL) as fp:
                self._release_info = json.loads(fp.read().decode("UTF-8"))

        threading.Thread(target=work, daemon=True).start()

    def _process_release_info(self):
        version_str = (
            self._release_info["tag_name"] + " (" + self._release_info["published_at"][:10] + ")"
        )
        self._version_label.configure(text=version_str)
        # self._install_button.configure(text=_("Install") + " " + version_str)

        candidates = [
            asset
            for asset in self._release_info["assets"]
            if asset["name"].endswith(".hex")
            and "micropython" in asset["name"].lower()
            and 400000 < asset["size"] < 800000
        ]
        if len(candidates) == 0:
            return self._close_with_error(
                "Could not find the right hex file from the release info (%s)" % LATEST_RELEASE_URL
            )
        elif len(candidates) > 1:
            return self._close_with_error(
                "Found several possible hex files from the release info (%s)" % LATEST_RELEASE_URL
            )
        else:
            self._hex_url = candidates[0]["browser_download_url"]
            self._hex_size = candidates[0]["size"]

    def _close_with_error(self, text):
        messagebox.showerror("Error", text)
        self._close()

    def _update_state(self):
        if self._state in ["closing", "closed"]:
            return

        if self._state == "starting" and self._release_info is not None:
            self._process_release_info()
            self._state = "ready"

        if self._state == "ready":
            self._install_button.state(["!disabled"])
        else:
            self._install_button.state(["disabled"])

        if self._state == "installing":
            self._progress_bar.configure(value=self._bytes_copied)
            self._old_bytes_copied = self._bytes_copied
            self._state_label.configure(text=_("Installing ..."))

        if self._state == "done":
            self._progress_bar.configure(value=0)
            self._state_label.configure(text=_("Done!") + " " + _("You can now close this dialog."))
            self._close_button.configure(text=_("Close"))

        if self._state != "done":
            self.after(200, self._update_state)

    def _start_installing(self):

        self._progress_bar.configure(maximum=self._hex_size)

        def work():
            self._copy_progess = 0

            target = os.path.join(self._target_dir, "micropython.hex")

            with urlopen(self._hex_url, timeout=5) as fsrc:
                with open(target, "wb") as fdst:
                    while True:
                        buf = fsrc.read(8 * 1024)
                        if not buf:
                            break

                        fdst.write(buf)
                        fdst.flush()
                        os.fsync(fdst)
                        self._bytes_copied += len(buf)

            self._state = "done"

        self._state = "installing"
        threading.Thread(target=work, daemon=True).start()

    def _close(self):
        if self._state == "installing" and not messagebox.askyesno(
            "Really cancel?", "Are you sure you want to cancel?"
        ):
            return

        self._state = "closing"
        self.destroy()
        self._state = "closed"


def load_plugin():
    add_micropython_backend(
        "microbit", MicrobitProxy, "MicroPython (BBC micro:bit)", MicrobitConfigPage
    )



import os
import ast
import sys
from thonny.backend import VM, get_vm


def augment_ast(root):
    mode = os.environ.get("PGZERO_MODE", "False")
    assert mode != "False"

    warning_prelude = "WARNING: Pygame Zero mode is turned on (Run → Pygame Zero mode)"
    try:
        import pgzero  # @UnusedImport
    except ImportError:
        if mode == "True":
            print(
                warning_prelude
                + ",\nbut pgzero module is not found. Running program in regular mode.\n",
                file=sys.stderr,
            )
        else:
            assert mode == "auto"

        return

    # Check if draw is defined
    for stmt in root.body:
        if isinstance(stmt, ast.FunctionDef) and stmt.name == "draw":
            break
    else:
        if mode == "auto":
            return
        else:
            print(
                warning_prelude
                + ",\nbut your program doesn't look like usual Pygame Zero program\n"
                + "(draw function is missing).\n",
                file=sys.stderr,
            )

    # need more checks in auto mode
    if mode == "auto":
        # check that draw method is not called in the code
        for node in ast.walk(root):
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Name)
                and node.func.id == "draw"
            ):
                return

    # prepend "import pgzrun as __pgzrun"
    imp = ast.Import([ast.alias("pgzrun", "__pgzrun")])
    imp.lineno = 0
    imp.col_offset = 0
    ast.fix_missing_locations(imp)
    imp.tags = {"ignore"}
    root.body.insert(0, imp)

    # append "__pgzrun.go()"
    go = ast.Expr(
        ast.Call(ast.Attribute(ast.Name("__pgzrun", ast.Load()), "go", ast.Load()), [], [])
    )
    go.lineno = 1000000
    go.col_offset = 0
    ast.fix_missing_locations(go)
    go.tags = {"ignore"}
    root.body.append(go)


def patched_editor_autocomplete(self, cmd):
    # Make extra builtins visible for Jedi
    prefix = "from pgzero.builtins import *\n"
    cmd["source"] = prefix + cmd["source"]
    cmd["row"] = cmd["row"] + 1

    result = get_vm()._original_editor_autocomplete(cmd)
    result["row"] = result["row"] - 1
    result["source"] = result["source"][len(prefix) :]

    return result


def load_plugin():
    if os.environ.get("PGZERO_MODE", "False").lower() == "false":
        return

    get_vm().add_ast_postprocessor(augment_ast)
    VM._original_editor_autocomplete = VM._cmd_editor_autocomplete
    VM._cmd_editor_autocomplete = patched_editor_autocomplete



class I2SOut:
    ""

    def deinit():
        pass

    def pause():
        pass

    paused = None

    def play():
        pass

    playing = None

    def resume():
        pass

    def stop():
        pass


class PDMIn:
    ""

    def deinit():
        pass

    def record():
        pass

    sample_rate = None



import tkinter as tk
from tkinter import messagebox

from thonny import get_runner, get_workbench
from thonny.codeview import CodeViewText
from thonny.common import InlineCommand
from thonny.shell import ShellText


# TODO: adjust the window position in cases where it's too close to bottom or right edge - but make sure the current line is shown
"""Completions get computed on the backend, therefore getting the completions is
asynchronous.
"""


class Completer(tk.Listbox):
    def __init__(self, text):
        tk.Listbox.__init__(
            self, master=text, font="SmallEditorFont", activestyle="dotbox", exportselection=False
        )

        self.text = text
        self.completions = []

        self.doc_label = tk.Label(
            master=text, text="Aaappiiiii", bg="#ffffe0", justify="left", anchor="nw"
        )

        # Auto indenter will eat up returns, therefore I need to raise the priority
        # of this binding
        self.text_priority_bindtag = "completable" + str(self.text.winfo_id())
        self.text.bindtags((self.text_priority_bindtag,) + self.text.bindtags())
        self.text.bind_class(self.text_priority_bindtag, "<Key>", self._on_text_keypress, True)

        self.text.bind("<<TextChange>>", self._on_text_change, True)  # Assuming TweakableText

        # for cases when Listbox gets focus
        self.bind("<Escape>", self._close)
        self.bind("<Return>", self._insert_current_selection)
        self.bind("<Double-Button-1>", self._insert_current_selection)
        self._bind_result_event()

    def _bind_result_event(self):
        # TODO: remove binding when editor gets closed
        get_workbench().bind("editor_autocomplete_response", self._handle_backend_response, True)

    def handle_autocomplete_request(self):
        row, column = self._get_position()
        source = self.text.get("1.0", "end-1c")
        get_runner().send_command(
            InlineCommand(
                "editor_autocomplete",
                source=source,
                row=row,
                column=column,
                filename=self._get_filename(),
            )
        )

    def _handle_backend_response(self, msg):
        row, column = self._get_position()
        source = self.text.get("1.0", "end-1c")

        if msg.source != source or msg.row != row or msg.column != column:
            # situation has changed, information is obsolete
            self._close()
        elif msg.get("error"):
            self._close()
            messagebox.showerror("Autocomplete error", msg.error)
        else:
            self._present_completions(msg.completions)

    def _present_completions(self, completions):
        self.completions = completions

        # broadcast logging info
        row, column = self._get_position()
        get_workbench().event_generate(
            "AutocompleteProposal",
            text_widget=self.text,
            row=row,
            column=column,
            proposal_count=len(completions),
        )

        # present
        if len(completions) == 0:
            self._close()
        elif len(completions) == 1:
            self._insert_completion(completions[0])  # insert the only completion
            self._close()
        else:
            self._show_box(completions)

    def _show_box(self, completions):
        self.delete(0, self.size())
        self.insert(
            0, *[c["name"] + ("=" if c["complete"].endswith("=") else "") for c in completions]
        )
        self.activate(0)
        self.selection_set(0)

        # place box
        if not self._is_visible():

            # _, _, _, list_box_height = self.bbox(0)
            height = 100  # min(150, list_box_height * len(completions) * 1.15)
            typed_name_length = len(completions[0]["name"]) - len(completions[0]["complete"])
            text_box_x, text_box_y, _, text_box_height = self.text.bbox(
                "insert-%dc" % typed_name_length
            )

            # should the box appear below or above cursor?
            space_below = self.master.winfo_height() - text_box_y - text_box_height
            space_above = text_box_y

            if space_below >= height or space_below > space_above:
                height = min(height, space_below)
                y = text_box_y + text_box_height
            else:
                height = min(height, space_above)
                y = text_box_y - height

            width = 400
            self.place(x=text_box_x, y=y, width=width, height=height)

            self._update_doc()

    def _update_doc(self):
        c = self._get_selected_completion()

        if c is None:
            self.doc_label["text"] = ""
            self.doc_label.place_forget()
        else:
            docstring = c.get("docstring", None)
            if docstring:
                self.doc_label["text"] = docstring
                self.doc_label.place(
                    x=self.winfo_x() + self.winfo_width(),
                    y=self.winfo_y(),
                    width=400,
                    height=self.winfo_height(),
                )
            else:
                self.doc_label["text"] = ""
                self.doc_label.place_forget()

    def _is_visible(self):
        return self.winfo_ismapped()

    def _insert_completion(self, completion):
        typed_len = len(completion["name"]) - len(completion["complete"].strip("="))
        typed_prefix = self.text.get("insert-{}c".format(typed_len), "insert")
        get_workbench().event_generate(
            "AutocompleteInsertion",
            text_widget=self.text,
            typed_prefix=typed_prefix,
            completed_name=completion["name"],
        )

        if self._is_visible():
            self._close()

        if not completion["name"].startswith(typed_prefix):
            # eg. case of the prefix was not correct
            self.text.delete("insert-{}c".format(typed_len), "insert")
            self.text.insert("insert", completion["name"])
        else:
            self.text.insert("insert", completion["complete"])

    def _get_filename(self):
        # TODO: allow completing in shell
        if not isinstance(self.text, CodeViewText):
            return None

        codeview = self.text.master

        editor = get_workbench().get_editor_notebook().get_current_editor()
        if editor.get_code_view() is codeview:
            return editor.get_filename()
        else:
            return None

    def _move_selection(self, delta):
        selected = self.curselection()
        if len(selected) == 0:
            index = 0
        else:
            index = selected[0]

        index += delta
        index = max(0, min(self.size() - 1, index))

        self.selection_clear(0, self.size() - 1)
        self.selection_set(index)
        self.activate(index)
        self.see(index)
        self._update_doc()

    def _get_request_id(self):
        return "autocomplete_" + str(self.text.winfo_id())

    def _get_position(self):
        return map(int, self.text.index("insert").split("."))

    def _on_text_keypress(self, event=None):
        if not self._is_visible():
            return None

        if event.keysym == "Escape":
            self._close()
            return "break"
        elif event.keysym in ["Up", "KP_Up"]:
            self._move_selection(-1)
            return "break"
        elif event.keysym in ["Down", "KP_Down"]:
            self._move_selection(1)
            return "break"
        elif event.keysym in ["Return", "KP_Enter", "Tab"]:
            assert self.size() > 0
            self._insert_current_selection()
            return "break"

        return None

    def _insert_current_selection(self, event=None):
        self._insert_completion(self._get_selected_completion())

    def _get_selected_completion(self):
        sel = self.curselection()
        if len(sel) != 1:
            return None

        return self.completions[sel[0]]

    def _on_text_change(self, event=None):
        if self._is_visible():
            self.handle_autocomplete_request()

    def _close(self, event=None):
        self.place_forget()
        self.doc_label.place_forget()
        self.text.focus_set()

    def on_text_click(self, event=None):
        if self._is_visible():
            self._close()


class ShellCompleter(Completer):
    def _bind_result_event(self):
        # TODO: remove binding when editor gets closed
        get_workbench().bind("shell_autocomplete_response", self._handle_backend_response, True)

    def handle_autocomplete_request(self):
        source = self._get_prefix()

        get_runner().send_command(InlineCommand("shell_autocomplete", source=source))

    def _handle_backend_response(self, msg):
        # check if the response is relevant for current state
        if msg.source != self._get_prefix():
            self._close()
        else:
            self._present_completions(msg.completions)

    def _get_prefix(self):
        return self.text.get("input_start", "insert")  # TODO: allow multiple line input


def handle_autocomplete_request(event=None):
    if event is None:
        text = get_workbench().focus_get()
    else:
        text = event.widget

    _handle_autocomplete_request_for_text(text)


def _handle_autocomplete_request_for_text(text):
    if not hasattr(text, "autocompleter"):
        if isinstance(text, (CodeViewText, ShellText)):
            if isinstance(text, CodeViewText):
                text.autocompleter = Completer(text)
            elif isinstance(text, ShellText):
                text.autocompleter = ShellCompleter(text)
            text.bind("<1>", text.autocompleter.on_text_click)
        else:
            return

    text.autocompleter.handle_autocomplete_request()


def patched_perform_midline_tab(text, event):
    if isinstance(text, ShellText):
        option_name = "edit.tab_complete_in_shell"
    else:
        option_name = "edit.tab_complete_in_editor"

    if get_workbench().get_option(option_name):
        if not text.has_selection():
            _handle_autocomplete_request_for_text(text)
            return "break"
        else:
            return None

    return text.perform_smart_tab(event)


def load_plugin() -> None:

    get_workbench().add_command(
        "autocomplete",
        "edit",
        _("Auto-complete"),
        handle_autocomplete_request,
        default_sequence="<Control-space>"
        # TODO: tester
    )

    get_workbench().set_default("edit.tab_complete_in_editor", True)
    get_workbench().set_default("edit.tab_complete_in_shell", True)

    CodeViewText.perform_midline_tab = patched_perform_midline_tab  # type: ignore
    ShellText.perform_midline_tab = patched_perform_midline_tab  # type: ignore



def chdir():
    pass


def getcwd():
    pass


def listdir():
    pass


def mkdir():
    pass


def remove():
    pass


def rename():
    pass


def rmdir():
    pass


sep = "/"


def stat():
    pass


def statvfs():
    pass


def sync():
    pass


def uname():
    pass


def unlink():
    pass


def urandom():
    pass



import ast
import datetime
import gzip
import json
import logging
import os.path
import sys
import tempfile
import textwrap
import tkinter as tk
import urllib.request
import webbrowser
from collections import namedtuple
from tkinter import messagebox, ttk
from typing import (
    Dict,  # pylint disable=unused-import
    Iterable,
    List,  # pylint disable=unused-import
    Optional,  # pylint disable=unused-import
    Tuple,  # pylint disable=unused-import
    Type,  # pylint disable=unused-import
    Union,  # pylint disable=unused-import
)


import thonny
from thonny import get_workbench, rst_utils, tktextext, ui_utils, get_runner
from thonny.common import ToplevelResponse, read_source
from thonny.misc_utils import levenshtein_damerau_distance, running_on_mac_os
from thonny.ui_utils import scrollbar_style, CommonDialog
from thonny.running import CPythonProxy
import subprocess

Suggestion = namedtuple("Suggestion", ["symbol", "title", "body", "relevance"])

_program_analyzer_classes = []  # type: List[Type[ProgramAnalyzer]]
_last_feedback_timestamps = {}  # type: Dict[str, str]
_error_helper_classes = {}  # type: Dict[str, List[Type[ErrorHelper]]]


class AssistantView(tktextext.TextFrame):
    def __init__(self, master):
        tktextext.TextFrame.__init__(
            self,
            master,
            text_class=AssistantRstText,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            read_only=True,
            wrap="word",
            font="TkDefaultFont",
            # cursor="arrow",
            padx=10,
            pady=0,
            insertwidth=0,
        )

        self._analyzer_instances = []

        self._snapshots_per_main_file = {}
        self._current_snapshot = None

        self._accepted_warning_sets = []

        self.text.tag_configure(
            "section_title",
            spacing3=5,
            font="BoldTkDefaultFont",
            # foreground=get_syntax_options_for_tag("stderr")["foreground"]
        )
        self.text.tag_configure(
            "intro",
            # font="ItalicTkDefaultFont",
            spacing3=10,
        )
        self.text.tag_configure("relevant_suggestion_title", font="BoldTkDefaultFont")
        self.text.tag_configure("suggestion_title", lmargin2=16, spacing1=5, spacing3=5)
        self.text.tag_configure("suggestion_body", lmargin1=16, lmargin2=16)
        self.text.tag_configure("body", font="ItalicTkDefaultFont")

        main_font = tk.font.nametofont("TkDefaultFont")

        # Underline on font looks better than underline on tag
        italic_underline_font = main_font.copy()
        italic_underline_font.configure(slant="italic", size=main_font.cget("size"), underline=True)

        self.text.tag_configure("feedback_link", justify="right", font=italic_underline_font)
        self.text.tag_bind("feedback_link", "<ButtonRelease-1>", self._ask_feedback, True)
        self.text.tag_configure("python_errors_link", justify="right", font=italic_underline_font)
        self.text.tag_bind(
            "python_errors_link",
            "<ButtonRelease-1>",
            lambda e: get_workbench().open_url("errors.rst"),
            True,
        )

        get_workbench().bind("ToplevelResponse", self.handle_toplevel_response, True)

        add_error_helper("*", GenericErrorHelper)

    def handle_toplevel_response(self, msg: ToplevelResponse) -> None:
        # Can be called by event system or by Workbench
        # (if Assistant wasn't created yet but an error came)
        if not msg.get("user_exception") and msg.get("command_name") in [
            "execute_system_command",
            "execute_source",
        ]:
            # Shell commands may be used to investigate the problem, don't clear assistance
            return

        self._clear()

        if not isinstance(get_runner().get_backend_proxy(), CPythonProxy):
            # TODO: add some support for MicroPython as well
            return

        # prepare for snapshot
        key = msg.get("filename", "<pyshell>")
        self._current_snapshot = {
            "timestamp": datetime.datetime.now().isoformat()[:19],
            "main_file_path": key,
        }
        self._snapshots_per_main_file.setdefault(key, [])
        self._snapshots_per_main_file[key].append(self._current_snapshot)

        if msg.get("user_exception"):
            if not msg["user_exception"].get("message", None):
                msg["user_exception"]["message"] = "<no message>"

            self._exception_info = msg["user_exception"]
            self._explain_exception(msg["user_exception"])
            if get_workbench().get_option("assistance.open_assistant_on_errors"):
                get_workbench().show_view("AssistantView", set_focus=False)
        else:
            self._exception_info = None

        if msg.get("filename") and os.path.exists(msg["filename"]):
            self.main_file_path = msg["filename"]
            source = read_source(msg["filename"])
            self._start_program_analyses(
                msg["filename"], source, _get_imported_user_files(msg["filename"], source)
            )
        else:
            self.main_file_path = None
            self._present_conclusion()

    def _explain_exception(self, error_info):
        rst = (
            self._get_rst_prelude()
            + rst_utils.create_title(
                error_info["type_name"] + ": " + rst_utils.escape(error_info["message"])
            )
            + "\n"
        )

        if (
            error_info.get("lineno") is not None
            and error_info.get("filename")
            and os.path.exists(error_info["filename"])
        ):
            rst += "`%s, line %d <%s>`__\n\n" % (
                os.path.basename(error_info["filename"]),
                error_info["lineno"],
                self._format_file_url(error_info),
            )

        helpers = [
            helper_class(error_info)
            for helper_class in (
                _error_helper_classes.get(error_info["type_name"], []) + _error_helper_classes["*"]
            )
        ]

        best_intro = helpers[0]
        for helper in helpers:
            if helper.intro_confidence > best_intro.intro_confidence:
                best_intro = helper

        # intro
        if best_intro.intro_text:
            rst += (
                ".. note::\n"
                + "    "
                + best_intro.intro_text.strip().replace("\n", "\n\n    ")
                + "\n\n"
            )

        suggestions = [
            suggestion
            for helper in helpers
            for suggestion in helper.suggestions
            if suggestion is not None
        ]
        suggestions = sorted(suggestions, key=lambda s: s.relevance, reverse=True)

        if suggestions[0].relevance > 1 or best_intro.intro_confidence > 1:
            relevance_threshold = 2
        else:
            # use relevance 1 only when there is nothing better
            relevance_threshold = 1

        suggestions = [s for s in suggestions if s.relevance >= relevance_threshold]

        for i, suggestion in enumerate(suggestions):
            rst += self._format_suggestion(
                suggestion,
                i == len(suggestions) - 1,
                # TODO: is it good if first is preopened?
                # It looks cleaner if it is not.
                False,  # i==0
            )

        self._current_snapshot["exception_suggestions"] = [
            dict(sug._asdict()) for sug in suggestions
        ]

        self.text.append_rst(rst)
        self._append_text("\n")

        self._current_snapshot["exception_type_name"] = error_info["type_name"]
        self._current_snapshot["exception_message"] = error_info["message"]
        self._current_snapshot["exception_file_path"] = error_info["filename"]
        self._current_snapshot["exception_lineno"] = error_info["lineno"]
        self._current_snapshot["exception_rst"] = rst  # for debugging purposes

    def _format_suggestion(self, suggestion, last, initially_open):
        return (
            # assuming that title is already in rst format
            ".. topic:: "
            + suggestion.title
            + "\n"
            + "    :class: toggle%s%s\n"
            % (", open" if initially_open else "", ", tight" if not last else "")
            + "    \n"
            + textwrap.indent(suggestion.body, "    ")
            + "\n\n"
        )

    def _append_text(self, chars, tags=()):
        self.text.direct_insert("end", chars, tags=tags)

    def _clear(self):
        self._accepted_warning_sets.clear()
        for wp in self._analyzer_instances:
            wp.cancel_analysis()
        self._analyzer_instances = []
        self.text.clear()

    def _start_program_analyses(self, main_file_path, main_file_source, imported_file_paths):

        for cls in _program_analyzer_classes:
            analyzer = cls(self._accept_warnings)
            if analyzer.is_enabled():
                self._analyzer_instances.append(analyzer)

        if not self._analyzer_instances:
            return

        self._append_text("\nAnalyzing your code ...", ("em",))

        # save snapshot of current source
        self._current_snapshot["main_file_path"] = main_file_path
        self._current_snapshot["main_file_source"] = main_file_source
        self._current_snapshot["imported_files"] = {
            name: read_source(name) for name in imported_file_paths
        }

        # start the analysis
        for analyzer in self._analyzer_instances:
            analyzer.start_analysis(main_file_path, imported_file_paths)

    def _accept_warnings(self, analyzer, warnings):
        if analyzer.cancelled:
            return

        self._accepted_warning_sets.append(warnings)
        if len(self._accepted_warning_sets) == len(self._analyzer_instances):
            self._present_warnings()
            self._present_conclusion()

    def _present_conclusion(self):

        if not self.text.get("1.0", "end").strip():
            if self.main_file_path is not None and os.path.exists(self.main_file_path):
                self._append_text("\n")
                self.text.append_rst(
                    "The code in `%s <%s>`__ looks good.\n\n"
                    % (
                        os.path.basename(self.main_file_path),
                        self._format_file_url({"filename": self.main_file_path}),
                    )
                )
                self.text.append_rst(
                    "If it is not working as it should, "
                    + "then consider using some general "
                    + "`debugging techniques <debugging.rst>`__.\n\n",
                    ("em",),
                )

        if self.text.get("1.0", "end").strip():
            self._append_feedback_link()

        if self._exception_info:
            self._append_text(
                "General advice on dealing with errors.\n", ("a", "python_errors_link")
            )

    def _present_warnings(self):
        warnings = [w for ws in self._accepted_warning_sets for w in ws]
        self.text.direct_delete("end-2l linestart", "end-1c lineend")

        if not warnings:
            return

        if self._exception_info is None:
            intro = "May be ignored if you are happy with your program."
        else:
            intro = "May help you find the cause of the error."

        rst = (
            self._get_rst_prelude()
            + rst_utils.create_title("Warnings")
            + ":remark:`%s`\n\n" % intro
        )

        by_file = {}
        for warning in warnings:
            if warning["filename"] not in by_file:
                by_file[warning["filename"]] = []
            if warning not in by_file[warning["filename"]]:
                # Pylint may give double warnings (eg. when module imports itself)
                by_file[warning["filename"]].append(warning)

        for filename in by_file:
            rst += "`%s <%s>`__\n\n" % (
                os.path.basename(filename),
                self._format_file_url(dict(filename=filename)),
            )
            file_warnings = sorted(
                by_file[filename], key=lambda x: (x.get("lineno", 0), -x.get("relevance", 1))
            )

            for i, warning in enumerate(file_warnings):
                rst += self._format_warning(warning, i == len(file_warnings) - 1) + "\n"

            rst += "\n"

        self.text.append_rst(rst)

        # save snapshot
        self._current_snapshot["warnings_rst"] = rst
        self._current_snapshot["warnings"] = warnings

        if get_workbench().get_option("assistance.open_assistant_on_warnings"):
            get_workbench().show_view("AssistantView")

    def _format_warning(self, warning, last):
        title = rst_utils.escape(warning["msg"].splitlines()[0])
        if warning.get("lineno") is not None:
            url = self._format_file_url(warning)
            if warning.get("lineno"):
                title = "`Line %d <%s>`__ : %s" % (warning["lineno"], url, title)

        if warning.get("explanation_rst"):
            explanation_rst = warning["explanation_rst"]
        elif warning.get("explanation"):
            explanation_rst = rst_utils.escape(warning["explanation"])
        else:
            explanation_rst = ""

        if warning.get("more_info_url"):
            explanation_rst += "\n\n`More info online <%s>`__" % warning["more_info_url"]

        explanation_rst = explanation_rst.strip()
        topic_class = "toggle" if explanation_rst else "empty"
        if not explanation_rst:
            explanation_rst = "n/a"

        return (
            ".. topic:: %s\n" % title
            + "    :class: "
            + topic_class
            + ("" if last else ", tight")
            + "\n"
            + "    \n"
            + textwrap.indent(explanation_rst, "    ")
            + "\n\n"
        )

    def _append_feedback_link(self):
        self._append_text("Was it helpful or confusing?\n", ("a", "feedback_link"))

    def _format_file_url(self, atts):
        return format_file_url(atts["filename"], atts.get("lineno"), atts.get("col_offset"))

    def _ask_feedback(self, event=None):

        all_snapshots = self._snapshots_per_main_file[self._current_snapshot["main_file_path"]]

        # TODO: select only snapshots which are not sent yet
        snapshots = all_snapshots

        ui_utils.show_dialog(FeedbackDialog(get_workbench(), self.main_file_path, snapshots))

    def _get_rst_prelude(self):
        return ".. default-role:: code\n\n" + ".. role:: light\n\n" + ".. role:: remark\n\n"


class AssistantRstText(rst_utils.RstText):
    def configure_tags(self):
        super().configure_tags()

        main_font = tk.font.nametofont("TkDefaultFont")

        italic_font = main_font.copy()
        italic_font.configure(slant="italic", size=main_font.cget("size"))

        h1_font = main_font.copy()
        h1_font.configure(weight="bold", size=main_font.cget("size"))

        self.tag_configure("h1", font=h1_font, spacing3=0, spacing1=10)
        self.tag_configure("topic_title", font="TkDefaultFont")

        self.tag_configure("topic_body", font=italic_font, spacing1=10, lmargin1=25, lmargin2=25)

        self.tag_raise("sel")


class Helper:
    def get_intro(self) -> Tuple[str, int]:
        raise NotImplementedError()

    def get_suggestions(self) -> Iterable[Suggestion]:
        raise NotImplementedError()


class ErrorHelper(Helper):
    def __init__(self, error_info):

        # TODO: don't repeat all this for all error helpers
        self.error_info = error_info

        self.last_frame = error_info["stack"][-1]
        self.last_frame_ast = None
        if self.last_frame.source:
            try:
                self.last_frame_ast = ast.parse(self.last_frame.source, self.last_frame.filename)
            except SyntaxError:
                pass

        self.last_frame_module_source = None
        self.last_frame_module_ast = None
        if self.last_frame.code_name == "<module>":
            self.last_frame_module_source = self.last_frame.source
            self.last_frame_module_ast = self.last_frame_ast
        elif self.last_frame.filename is not None:
            try:
                self.last_frame_module_source = read_source(self.last_frame.filename)
                self.last_frame_module_ast = ast.parse(self.last_frame_module_source)
            except Exception:
                pass

        self.intro_confidence = 1
        self.intro_text = ""
        self.suggestions = []


class GenericErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)

        self.intro_text = "No specific suggestions for this error (yet)."
        self.intro_confidence = 1
        self.suggestions = [
            Suggestion(
                "ask-for-specific-support",
                "Let Thonny developers know",
                "Click on the feedback link at the bottom of this panel to let Thonny developers know "
                + "about your problem. They may add support for "
                + "such cases in future Thonny versions.",
                1,
            )
        ]

        if error_info["message"].lower() != "invalid syntax":
            self.suggestions.append(
                Suggestion(
                    "generic-search-the-web",
                    "Search the web",
                    "Try performing a web search for\n\n``Python %s: %s``"
                    % (self.error_info["type_name"], rst_utils.escape(self.error_info["message"])),
                    1,
                )
            )


class ProgramAnalyzer:
    def __init__(self, on_completion):
        self.completion_handler = on_completion
        self.cancelled = False

    def is_enabled(self):
        return True

    def start_analysis(self, main_file_path, imported_file_paths):
        raise NotImplementedError()

    def cancel_analysis(self):
        pass


class SubprocessProgramAnalyzer(ProgramAnalyzer):
    def __init__(self, on_completion):
        super().__init__(on_completion)
        self._proc = None

    def cancel_analysis(self):
        self.cancelled = True
        if self._proc is not None:
            self._proc.kill()


class LibraryErrorHelper(ErrorHelper):
    """Explains exceptions, which doesn't happen in user code"""

    def get_intro(self):
        return "This error happened in library code. This may mean a bug in "

    def get_suggestions(self):
        return []


class FeedbackDialog(CommonDialog):
    def __init__(self, master, main_file_path, all_snapshots):
        super().__init__(master=master)
        main_frame = ttk.Frame(self)
        main_frame.grid(row=0, column=0, sticky="nsew")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.main_file_path = main_file_path
        self.snapshots = self._select_unsent_snapshots(all_snapshots)

        self.title("Send feedback for Assistant")

        padx = 15

        intro_label = ttk.Label(
            main_frame,
            text="Below are the messages Assistant gave you in response to "
            + (
                "using the shell"
                if self._happened_in_shell()
                else "testing '" + os.path.basename(main_file_path) + "'"
            )
            + " since "
            + self._get_since_str()
            + ".\n\n"
            + "In order to improve this feature, Thonny developers would love to know how "
            + "useful or confusing these messages were. We will only collect version "
            + "information and the data you enter or approve on this form.",
            wraplength=550,
        )
        intro_label.grid(row=1, column=0, columnspan=3, sticky="nw", padx=padx, pady=(15, 15))

        tree_label = ttk.Label(
            main_frame,
            text="Which messages were helpful (H) or confusing (C)?       Click on  [  ]  to mark!",
        )
        tree_label.grid(row=2, column=0, columnspan=3, sticky="nw", padx=padx, pady=(15, 0))
        tree_frame = ui_utils.TreeFrame(
            main_frame,
            columns=["helpful", "confusing", "title", "group", "symbol"],
            displaycolumns=["helpful", "confusing", "title"],
            height=10,
            borderwidth=1,
            relief="groove",
        )
        tree_frame.grid(row=3, column=0, columnspan=3, sticky="nsew", padx=padx)
        self.tree = tree_frame.tree
        self.tree.column("helpful", width=30, anchor=tk.CENTER, stretch=False)
        self.tree.column("confusing", width=30, anchor=tk.CENTER, stretch=False)
        self.tree.column("title", width=350, anchor=tk.W, stretch=True)

        self.tree.heading("helpful", text="H", anchor=tk.CENTER)
        self.tree.heading("confusing", text="C", anchor=tk.CENTER)
        self.tree.heading("title", text="Group / Message", anchor=tk.W)
        self.tree["show"] = ("headings",)
        self.tree.bind("<1>", self._on_tree_click, True)
        main_font = tk.font.nametofont("TkDefaultFont")
        bold_font = main_font.copy()
        bold_font.configure(weight="bold", size=main_font.cget("size"))
        self.tree.tag_configure("group", font=bold_font)

        self.include_thonny_id_var = tk.IntVar(value=1)
        include_thonny_id_check = ttk.Checkbutton(
            main_frame,
            variable=self.include_thonny_id_var,
            onvalue=1,
            offvalue=0,
            text="Include Thonny's installation time (allows us to group your submissions)",
        )
        include_thonny_id_check.grid(
            row=4, column=0, columnspan=3, sticky="nw", padx=padx, pady=(5, 0)
        )

        self.include_snapshots_var = tk.IntVar(value=1)
        include_snapshots_check = ttk.Checkbutton(
            main_frame,
            variable=self.include_snapshots_var,
            onvalue=1,
            offvalue=0,
            text="Include snapshots of the code and Assistant responses at each run",
        )
        include_snapshots_check.grid(
            row=5, column=0, columnspan=3, sticky="nw", padx=padx, pady=(0, 0)
        )

        comments_label = ttk.Label(main_frame, text="Any comments? Enhancement ideas?")
        comments_label.grid(row=6, column=0, columnspan=3, sticky="nw", padx=padx, pady=(15, 0))
        self.comments_text_frame = tktextext.TextFrame(
            main_frame,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            wrap="word",
            font="TkDefaultFont",
            # cursor="arrow",
            padx=5,
            pady=5,
            height=4,
            borderwidth=1,
            relief="groove",
        )
        self.comments_text_frame.grid(row=7, column=0, columnspan=3, sticky="nsew", padx=padx)

        url_font = tk.font.nametofont("TkDefaultFont").copy()
        url_font.configure(underline=1, size=url_font.cget("size"))
        preview_link = ttk.Label(
            main_frame,
            text="(Preview the data to be sent)",
            style="Url.TLabel",
            cursor="hand2",
            font=url_font,
        )
        preview_link.bind("<1>", self._preview_submission_data, True)
        preview_link.grid(row=8, column=0, sticky="nw", padx=15, pady=15)

        submit_button = ttk.Button(main_frame, text="Submit", width=10, command=self._submit_data)
        submit_button.grid(row=8, column=0, sticky="ne", padx=0, pady=15)

        cancel_button = ttk.Button(main_frame, text="Cancel", width=7, command=self._close)
        cancel_button.grid(row=8, column=1, sticky="ne", padx=(10, 15), pady=15)

        self.protocol("WM_DELETE_WINDOW", self._close)
        self.bind("<Escape>", self._close, True)

        main_frame.columnconfigure(0, weight=1)
        main_frame.rowconfigure(3, weight=3)
        main_frame.rowconfigure(6, weight=2)

        self._empty_box = "[  ]"
        self._checked_box = "[X]"
        self._populate_tree()

    def _happened_in_shell(self):
        return self.main_file_path is None or self.main_file_path.lower() == "<pyshell>"

    def _populate_tree(self):
        groups = {}

        for snap in self.snapshots:
            if snap.get("exception_message") and snap.get("exception_suggestions"):
                group = snap["exception_type_name"]
                groups.setdefault(group, set())
                for sug in snap["exception_suggestions"]:
                    groups[group].add((sug["symbol"], sug["title"]))

            # warnings group
            if snap.get("warnings"):
                group = "Warnings"
                groups.setdefault(group, set())
                for w in snap["warnings"]:
                    groups[group].add((w["symbol"], w["msg"]))

        for group in sorted(groups.keys(), key=lambda x: x.replace("Warnings", "z")):
            group_id = self.tree.insert("", "end", open=True, tags=("group",))
            self.tree.set(group_id, "title", group)

            for symbol, title in sorted(groups[group], key=lambda m: m[1]):
                item_id = self.tree.insert("", "end")
                self.tree.set(item_id, "helpful", self._empty_box)
                self.tree.set(item_id, "confusing", self._empty_box)
                self.tree.set(item_id, "title", title)
                self.tree.set(item_id, "symbol", symbol)
                self.tree.set(item_id, "group", group)

        self.tree.see("")

    def _on_tree_click(self, event):
        item_id = self.tree.identify("item", event.x, event.y)
        column = self.tree.identify_column(event.x)

        if not item_id or not column:
            return

        value_index = int(column[1:]) - 1
        values = list(self.tree.item(item_id, "values"))

        if values[value_index] == self._empty_box:
            values[value_index] = self._checked_box
        elif values[value_index] == self._checked_box:
            values[value_index] = self._empty_box
        else:
            return

        # update values
        self.tree.item(item_id, values=tuple(values))

    def _preview_submission_data(self, event=None):
        temp_path = os.path.join(
            tempfile.mkdtemp(),
            "ThonnyAssistantFeedback_"
            + datetime.datetime.now().isoformat().replace(":", ".")[:19]
            + ".txt",
        )
        data = self._collect_submission_data()
        with open(temp_path, "w", encoding="ascii") as fp:
            fp.write(data)

        if running_on_mac_os():
            subprocess.Popen(["open", "-e", temp_path])
        else:
            webbrowser.open(temp_path)

    def _collect_submission_data(self):
        tree_data = []

        for iid in self.tree.get_children():
            values = self.tree.item(iid, "values")
            tree_data.append(
                {
                    "helpful": values[0] == self._checked_box,
                    "confusing": values[1] == self._checked_box,
                    "message": values[2],
                    "group": values[3],
                    "symbol": values[4],
                }
            )

        submission = {
            "feedback_format_version": 1,
            "thonny_version": thonny.get_version(),
            "python_version": ".".join(map(str, sys.version_info[:3])),
            "message_feedback": tree_data,
            "comments": self.comments_text_frame.text.get("1.0", "end"),
        }

        try:
            import mypy.version

            submission["mypy_version"] = str(mypy.version.__version__)
        except ImportError:
            logging.exception("Could not get MyPy version")

        try:
            import pylint

            submission["pylint_version"] = str(pylint.__version__)
        except ImportError:
            logging.exception("Could not get Pylint version")

        if self.include_snapshots_var.get():
            submission["snapshots"] = self.snapshots

        if self.include_thonny_id_var.get():
            submission["thonny_timestamp"] = get_workbench().get_option(
                "general.configuration_creation_timestamp"
            )

        return json.dumps(submission, indent=2)

    def _submit_data(self):
        json_data = self._collect_submission_data()
        compressed_data = gzip.compress(json_data.encode("ascii"))

        def do_work():
            try:
                handle = urllib.request.urlopen(
                    "https://thonny.org/store_assistant_feedback.php",
                    data=compressed_data,
                    timeout=10,
                )
                return handle.read()
            except Exception as e:
                return str(e)

        result = ui_utils.run_with_waiting_dialog(self, do_work, description="Uploading")
        if result == b"OK":
            if self.snapshots:
                last_timestamp = self.snapshots[-1]["timestamp"]
                _last_feedback_timestamps[self.main_file_path] = last_timestamp
            messagebox.showinfo(
                "Done!",
                "Thank you for the feedback!\n\nLet us know again when Assistant\nhelps or confuses you!",
            )
            self._close()
        else:
            messagebox.showerror(
                "Problem",
                "Something went wrong:\n%s\n\nIf you don't mind, then try again later!"
                % result[:1000],
            )

    def _select_unsent_snapshots(self, all_snapshots):
        if self.main_file_path not in _last_feedback_timestamps:
            return all_snapshots
        else:
            return [
                s
                for s in all_snapshots
                if s["timestamp"] > _last_feedback_timestamps[self.main_file_path]
            ]

    def _close(self, event=None):
        self.destroy()

    def _get_since_str(self):
        if not self.snapshots:
            assert self.main_file_path in _last_feedback_timestamps
            since = datetime.datetime.strptime(
                _last_feedback_timestamps[self.main_file_path], "%Y-%m-%dT%H:%M:%S"
            )
        else:
            since = datetime.datetime.strptime(self.snapshots[0]["timestamp"], "%Y-%m-%dT%H:%M:%S")

        if since.date() == datetime.date.today() or (
            datetime.datetime.now() - since
        ) <= datetime.timedelta(hours=5):
            since_str = since.strftime("%X")
        else:
            # date and time without yer
            since_str = since.strftime("%c").replace(str(datetime.date.today().year), "")

        # remove seconds
        if since_str.count(":") == 2:
            i = since_str.rfind(":")
            if (
                i > 0
                and len(since_str[i + 1 : i + 3]) == 2
                and since_str[i + 1 : i + 3].isnumeric()
            ):
                since_str = since_str[:i] + since_str[i + 3 :]

        return since_str.strip()


def name_similarity(a, b):
    # TODO: tweak the result values
    a = a.replace("_", "")
    b = b.replace("_", "")

    minlen = min(len(a), len(b))

    if a.replace("0", "O").replace("1", "l") == b.replace("0", "O").replace("1", "l"):
        if minlen >= 4:
            return 7
        else:
            return 6

    a = a.lower()
    b = b.lower()

    if a == b:
        if minlen >= 4:
            return 7
        else:
            return 6

    if minlen <= 2:
        return 0

    # if names differ at final isolated digits,
    # then they are probably different vars, even if their
    # distance is small (eg. location_1 and location_2)
    if a[-1].isdigit() and not a[-2].isdigit() and b[-1].isdigit() and not b[-2].isdigit():
        return 0

    # same thing with _ + single char suffixes
    # (eg. location_a and location_b)
    if a[-2] == "_" and b[-2] == "_":
        return 0

    distance = levenshtein_damerau_distance(a, b, 5)

    if minlen <= 5:
        return max(8 - distance * 2, 0)
    elif minlen <= 10:
        return max(9 - distance * 2, 0)
    else:
        return max(10 - distance * 2, 0)


def _get_imported_user_files(main_file, source=None):
    assert os.path.isabs(main_file)

    if source is None:
        source = read_source(main_file)

    try:
        root = ast.parse(source, main_file)
    except SyntaxError:
        return set()

    main_dir = os.path.dirname(main_file)
    module_names = set()
    # TODO: at the moment only considers non-package modules
    for node in ast.walk(root):
        if isinstance(node, ast.Import):
            for item in node.names:
                module_names.add(item.name)
        elif isinstance(node, ast.ImportFrom):
            module_names.add(node.module)

    imported_files = set()

    for file in {name + ext for ext in [".py", ".pyw"] for name in module_names}:
        possible_path = os.path.join(main_dir, file)
        if os.path.exists(possible_path):
            imported_files.add(possible_path)

    return imported_files
    # TODO: add recursion


def add_program_analyzer(cls):
    _program_analyzer_classes.append(cls)


def add_error_helper(error_type_name, helper_class):
    _error_helper_classes.setdefault(error_type_name, [])
    _error_helper_classes[error_type_name].append(helper_class)


def format_file_url(filename, lineno, col_offset):
    s = "thonny-editor://" + rst_utils.escape(filename).replace(" ", "%20")
    if lineno is not None:
        s += "#" + str(lineno)
        if col_offset is not None:
            s += ":" + str(col_offset)

    return s


def init():
    get_workbench().set_default("assistance.open_assistant_on_errors", True)
    get_workbench().set_default("assistance.open_assistant_on_warnings", False)
    get_workbench().set_default("assistance.disabled_checks", [])
    get_workbench().add_view(AssistantView, _("Assistant"), "se", visible_by_default=False)



import tkinter as tk
from tkinter import ttk

from thonny import get_workbench, ui_utils
import traceback
from thonny.ui_utils import ems_to_pixels, CommonDialog


class ConfigurationDialog(CommonDialog):
    last_shown_tab_index = 0

    def __init__(self, master, page_records):
        super().__init__(master)
        width = ems_to_pixels(53)
        height = ems_to_pixels(43)
        self.geometry("%dx%d" % (width, height))
        self.title(_("Thonny options"))

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        main_frame = ttk.Frame(self)  # otherwise there is wrong color background with clam
        main_frame.grid(row=0, column=0, sticky=tk.NSEW)
        main_frame.columnconfigure(0, weight=1)
        main_frame.rowconfigure(0, weight=1)

        self._notebook = ttk.Notebook(main_frame)
        self._notebook.grid(row=0, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10)

        self._ok_button = ttk.Button(main_frame, text=_("OK"), command=self._ok, default="active")
        self._cancel_button = ttk.Button(main_frame, text=_("Cancel"), command=self._cancel)
        self._ok_button.grid(row=1, column=1, padx=(0, 11), pady=(0, 10))
        self._cancel_button.grid(row=1, column=2, padx=(0, 11), pady=(0, 10))

        self._page_records = []
        for key, title, page_class, order in sorted(page_records, key=lambda r: (r[3], r[0])):
            try:
                spacer = ttk.Frame(self)
                spacer.rowconfigure(0, weight=1)
                spacer.columnconfigure(0, weight=1)
                page = page_class(spacer)
                page.key = key
                self._page_records.append((key, title, page))
                page.grid(sticky=tk.NSEW, pady=(15, 10), padx=15)
                self._notebook.add(spacer, text=title)
            except Exception:
                traceback.print_exc()

        self.bind("<Return>", self._ok, True)
        self.bind("<Escape>", self._cancel, True)

        self._notebook.select(self._notebook.tabs()[ConfigurationDialog.last_shown_tab_index])

    def select_page(self, key):
        for i, tab in enumerate(self._notebook.tabs()):
            if self._page_records[i][0] == key:
                self._notebook.select(tab)

    def _ok(self, event=None):
        for key, title, page in self._page_records:
            try:
                if page.apply() == False:
                    return
            except Exception:
                get_workbench().report_exception("Error when applying options in " + title)

        self.destroy()

    def _cancel(self, event=None):
        for key, title, page in self._page_records:
            try:
                page.cancel()
            except Exception:
                get_workbench().report_exception("Error when cancelling options in " + title)

        self.destroy()

    def destroy(self):
        ConfigurationDialog.last_shown_tab_index = self._notebook.index(self._notebook.select())
        super().destroy()


class ConfigurationPage(ttk.Frame):
    """This is an example dummy implementation of a configuration page.
    
    It's not required that configuration pages inherit from this class
    (can be any widget), but the class must have constructor with single parameter
    for getting the master."""

    def __init__(self, master):
        ttk.Frame.__init__(self, master)

    def add_checkbox(
        self, flag_name, description, row=None, column=0, padx=0, pady=0, columnspan=1, tooltip=None
    ):
        variable = get_workbench().get_variable(flag_name)
        checkbox = ttk.Checkbutton(self, text=description, variable=variable)
        checkbox.grid(
            row=row, column=column, sticky=tk.W, padx=padx, pady=pady, columnspan=columnspan
        )

        if tooltip is not None:
            ui_utils.create_tooltip(checkbox, tooltip)

    def add_combobox(
        self, option_name, values, row=None, column=0, padx=0, pady=0, columnspan=1, width=None
    ):
        variable = get_workbench().get_variable(option_name)
        combobox = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=variable,
            state="readonly",
            height=15,
            width=width,
            values=values,
        )
        combobox.grid(
            row=row, column=column, sticky=tk.W, pady=pady, padx=padx, columnspan=columnspan
        )

    def add_entry(self, option_name, row=None, column=0, pady=0, padx=0, columnspan=1, **kw):
        variable = get_workbench().get_variable(option_name)
        entry = ttk.Entry(self, textvariable=variable, **kw)
        entry.grid(row=row, column=column, sticky=tk.W, pady=pady, columnspan=columnspan, padx=padx)

    def apply(self):
        """Apply method should return False, when page contains invalid
        input and configuration dialog should not be closed."""
        pass

    def cancel(self):
        """Called when dialog gets cancelled"""
        pass



# -*- coding: utf-8 -*-

import ast
import tkinter as tk
import traceback

from thonny import ast_utils, get_workbench, ui_utils
from thonny.common import TextRange, range_contains_smaller
import sys


class AstView(ui_utils.TreeFrame):
    def __init__(self, master):
        ui_utils.TreeFrame.__init__(
            self,
            master,
            columns=("range", "lineno", "col_offset", "end_lineno", "end_col_offset"),
            displaycolumns=(0,),
        )

        self._current_code_view = None
        self.tree.bind("<<TreeviewSelect>>", self._locate_code)
        self.tree.bind("<<Copy>>", self._copy_to_clipboard)
        self.tree.bind("<Map>", self._update, True)
        get_workbench().get_editor_notebook().bind("<<NotebookTabChanged>>", self._update)
        get_workbench().bind("Save", self._update, True)
        get_workbench().bind("SaveAs", self._update, True)
        get_workbench().bind_class("Text", "<Double-Button-1>", self._update, True)

        self.tree.column("#0", width=550, anchor=tk.W)
        self.tree.column("range", width=100, anchor=tk.W)
        self.tree.column("lineno", width=30, anchor=tk.W)
        self.tree.column("col_offset", width=30, anchor=tk.W)
        self.tree.column("end_lineno", width=30, anchor=tk.W)
        self.tree.column("end_col_offset", width=30, anchor=tk.W)

        self.tree.heading("#0", text="Node", anchor=tk.W)
        self.tree.heading("range", text="Code range", anchor=tk.W)

        self.tree["show"] = ("headings", "tree")
        self._current_source = None

        self._update(None)

    def _update(self, event):
        if not self.winfo_ismapped():
            return

        editor = get_workbench().get_editor_notebook().get_current_editor()

        if not editor:
            self._current_code_view = None
            return

        new_cw = editor.get_code_view()
        new_source = new_cw.get_content_as_bytes()
        if self._current_code_view == new_cw and self._current_source == new_source:
            return

        self._current_code_view = new_cw
        self._current_source = new_source
        selection = self._current_code_view.get_selected_range()

        self._clear_tree()

        if not self._current_source.strip():
            return

        try:
            root = ast_utils.parse_source(self._current_source)
            selected_ast_node = _find_closest_containing_node(root, selection)

        except Exception as e:
            self.tree.insert("", "end", text=str(e), open=True)
            traceback.print_exc()
            return

        def _format(key, node, parent_id):

            if isinstance(node, ast.AST):
                fields = [(key, val) for key, val in ast.iter_fields(node)]

                value_label = node.__class__.__name__

            elif isinstance(node, list):
                fields = list(enumerate(node))
                if len(node) == 0:
                    value_label = "[]"
                else:
                    value_label = "[...]"
            else:
                fields = []
                value_label = repr(node)

            item_text = str(key) + "=" + value_label
            node_id = self.tree.insert(parent_id, "end", text=item_text, open=True)
            if node == selected_ast_node:
                self.tree.see(node_id)
                self.tree.selection_add(node_id)

            if hasattr(node, "lineno") and hasattr(node, "col_offset"):
                self.tree.set(node_id, "lineno", node.lineno)
                self.tree.set(node_id, "col_offset", node.col_offset)

                range_str = str(node.lineno) + "." + str(node.col_offset)
                if hasattr(node, "end_lineno") and hasattr(node, "end_col_offset"):
                    self.tree.set(node_id, "end_lineno", node.end_lineno)
                    self.tree.set(node_id, "end_col_offset", node.end_col_offset)
                    range_str += "  -  " + str(node.end_lineno) + "." + str(node.end_col_offset)
                else:
                    # fallback
                    self.tree.set(node_id, "end_lineno", node.lineno)
                    self.tree.set(node_id, "end_col_offset", node.col_offset + 1)

                self.tree.set(node_id, "range", range_str)

            for field_key, field_value in fields:
                _format(field_key, field_value, node_id)

        _format("root", root, "")

    def _locate_code(self, event):
        if self._current_code_view is None:
            return

        iid = self.tree.focus()

        if iid != "":
            values = self.tree.item(iid)["values"]
            if isinstance(values, list) and len(values) >= 5:
                start_line, start_col, end_line, end_col = values[1:5]
                self._current_code_view.select_range(
                    TextRange(start_line, start_col, end_line, end_col)
                )

    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)

    def _copy_to_clipboard(self, event):
        self.clipboard_clear()
        if self._current_source is not None:
            pretty_ast = pretty(ast_utils.parse_source(self._current_source))
            self.clipboard_append(pretty_ast)


def _find_closest_containing_node(tree, text_range):
    # first look among children
    for child in ast.iter_child_nodes(tree):
        result = _find_closest_containing_node(child, text_range)
        if result is not None:
            return result

    # no suitable child was found
    if hasattr(tree, "lineno") and range_contains_smaller(
        TextRange(tree.lineno, tree.col_offset, tree.end_lineno, tree.end_col_offset), text_range
    ):
        return tree
    # nope
    else:
        return None


def pretty(node, key="/", level=0):
    """Used for exporting ASTView content"""
    if isinstance(node, ast.AST):
        fields = [(key, val) for key, val in ast.iter_fields(node)]
        value_label = node.__class__.__name__
        if isinstance(node, ast.Call):
            # Try to make 3.4 AST-s more similar to 3.5
            if sys.version_info[:2] == (3, 4):
                if ("kwargs", None) in fields:
                    fields.remove(("kwargs", None))
                if ("starargs", None) in fields:
                    fields.remove(("starargs", None))

            # TODO: translate also non-None kwargs and starargs

    elif isinstance(node, list):
        fields = list(enumerate(node))
        if len(node) == 0:
            value_label = "[]"
        else:
            value_label = "[...]"
    else:
        fields = []
        value_label = repr(node)

    item_text = level * "    " + str(key) + "=" + value_label

    if hasattr(node, "lineno"):
        item_text += " @ " + str(getattr(node, "lineno"))
        if hasattr(node, "col_offset"):
            item_text += "." + str(getattr(node, "col_offset"))

        if hasattr(node, "end_lineno"):
            item_text += "  -  " + str(getattr(node, "end_lineno"))
            if hasattr(node, "end_col_offset"):
                item_text += "." + str(getattr(node, "end_col_offset"))

    lines = [item_text] + [
        pretty(field_value, field_key, level + 1) for field_key, field_value in fields
    ]

    return "\n".join(lines)


def load_plugin() -> None:
    get_workbench().add_view(AstView, _("Program tree"), "s")



class Image:
    ""


accelerometer = None
button_a = None
button_b = None
compass = None
display = None
i2c = None


def panic():
    pass


pin0 = None
pin1 = None
pin10 = None
pin11 = None
pin12 = None
pin13 = None
pin14 = None
pin15 = None
pin16 = None
pin19 = None
pin2 = None
pin20 = None
pin3 = None
pin4 = None
pin5 = None
pin6 = None
pin7 = None
pin8 = None
pin9 = None


def reset():
    pass


def running_time():
    pass


def sleep():
    pass


spi = None


def temperature():
    pass


uart = None



def localtime():
    pass


def mktime():
    pass


def monotonic():
    pass


def monotonic_ns():
    pass


def sleep():
    pass


class struct_time:
    ""

    def count():
        pass

    def index():
        pass


def time():
    pass



import logging
import tkinter as tk
import traceback

from jedi import Script

from thonny import get_workbench, jedi_utils

tree = jedi_utils.import_python_tree()


class BaseNameHighlighter:
    def __init__(self, text):
        self.text = text
        self._update_scheduled = False

    def get_positions_for(self, source, line, column):
        raise NotImplementedError()

    def get_positions(self):
        index = self.text.index("insert")

        # ignore if cursor in open string
        if self.text.tag_prevrange("open_string", index) or self.text.tag_prevrange(
            "open_string3", index
        ):

            return set()

        source = self.text.get("1.0", "end")
        index_parts = index.split(".")
        line, column = int(index_parts[0]), int(index_parts[1])

        return self.get_positions_for(source, line, column)

    def schedule_update(self):
        def perform_update():
            try:
                self.update()
            finally:
                self._update_scheduled = False

        if not self._update_scheduled:
            self._update_scheduled = True
            self.text.after_idle(perform_update)

    def update(self):
        self.text.tag_remove("matched_name", "1.0", "end")

        if get_workbench().get_option("view.name_highlighting"):
            try:
                positions = self.get_positions()
                if len(positions) > 1:
                    for pos in positions:
                        start_index, end_index = pos[0], pos[1]
                        self.text.tag_add("matched_name", start_index, end_index)
            except Exception:
                logging.exception("Problem when updating name highlighting")


class VariablesHighlighter(BaseNameHighlighter):
    """This is heavy, but more correct solution for variables, than Script.usages provides 
    (at least for Jedi 0.10)"""

    def _is_name_function_call_name(self, name):
        stmt = name.get_definition()
        return stmt.type == "power" and stmt.children[0] == name

    def _is_name_function_definition(self, name):
        scope = name.get_definition()
        return (
            isinstance(scope, tree.Function)
            and hasattr(scope.children[1], "value")
            and scope.children[1].value == name.value
        )

    def _get_def_from_function_params(self, func_node, name):
        params = jedi_utils.get_params(func_node)
        for param in params:
            if param.children[0].value == name.value:
                return param.children[0]
        return None

    # copied from jedi's tree.py with a few modifications
    def _get_statement_for_position(self, node, pos):
        for c in node.children:
            # sorted here, because the end_pos property depends on the last child having the last position,
            # there seems to be a problem with jedi, where the children of a node are not always in the right order
            if isinstance(c, tree.Class):
                c.children.sort(key=lambda x: x.end_pos)
            if c.start_pos <= pos <= c.end_pos:
                if c.type not in ("decorated", "simple_stmt", "suite") and not isinstance(
                    c, (tree.Flow, tree.ClassOrFunc)
                ):
                    return c
                else:
                    try:
                        return jedi_utils.get_statement_of_position(c, pos)
                    except AttributeError:
                        traceback.print_exc()
        return None

    def _is_global_stmt_with_name(self, node, name_str):
        return (
            isinstance(node, tree.BaseNode)
            and node.type == "simple_stmt"
            and isinstance(node.children[0], tree.GlobalStmt)
            and node.children[0].children[1].value == name_str
        )

    def _find_definition(self, scope, name):

        # if the name is the name of a function definition
        if isinstance(scope, tree.Function):
            if scope.children[1] == name:
                return scope.children[1]  # 0th child is keyword "def", 1st is name
            else:
                definition = self._get_def_from_function_params(scope, name)
                if definition:
                    return definition

        for c in scope.children:
            if (
                isinstance(c, tree.BaseNode)
                and c.type == "simple_stmt"
                and isinstance(c.children[0], tree.ImportName)
            ):
                for n in c.children[0].get_defined_names():
                    if n.value == name.value:
                        return n
                # print(c.path_for_name(name.value))
            if (
                isinstance(c, tree.Function)
                and c.children[1].value == name.value
                and not isinstance(jedi_utils.get_parent_scope(c), tree.Class)
            ):
                return c.children[1]
            if isinstance(c, tree.BaseNode) and c.type == "suite":
                for x in c.children:
                    if self._is_global_stmt_with_name(x, name.value):
                        return self._find_definition(jedi_utils.get_parent_scope(scope), name)
                    if isinstance(x, tree.Name) and x.is_definition() and x.value == name.value:
                        return x
                    def_candidate = self._find_def_in_simple_node(x, name)
                    if def_candidate:
                        return def_candidate

        if not isinstance(scope, tree.Module):
            return self._find_definition(jedi_utils.get_parent_scope(scope), name)

        # if name itself is the left side of an assignment statement, then the name is the definition
        if name.is_definition():
            return name

        return None

    def _find_def_in_simple_node(self, node, name):
        if isinstance(node, tree.Name) and node.is_definition() and node.value == name.value:
            return name
        if not isinstance(node, tree.BaseNode):
            return None
        for c in node.children:
            return self._find_def_in_simple_node(c, name)

    def _get_dot_names(self, stmt):
        try:
            if stmt.children[1].children[0].value == ".":
                return stmt.children[0], stmt.children[1].children[1]
        except Exception:
            return ()
        return ()

    def _find_usages(self, name, stmt):
        # check if stmt is dot qualified, disregard these
        dot_names = self._get_dot_names(stmt)
        if len(dot_names) > 1 and dot_names[1].value == name.value:
            return set()

        # search for definition
        definition = self._find_definition(jedi_utils.get_parent_scope(name), name)

        searched_scopes = set()

        is_function_definition = (
            self._is_name_function_definition(definition) if definition else False
        )

        def find_usages_in_node(node, global_encountered=False):
            names = []
            if isinstance(node, tree.BaseNode):
                if jedi_utils.is_scope(node):
                    global_encountered = False
                    if node in searched_scopes:
                        return names
                    searched_scopes.add(node)
                    if isinstance(node, tree.Function):
                        d = self._get_def_from_function_params(node, name)
                        if d and d != definition:
                            return []

                for c in node.children:
                    dot_names = self._get_dot_names(c)
                    if len(dot_names) > 1 and dot_names[1].value == name.value:
                        continue
                    sub_result = find_usages_in_node(c, global_encountered=global_encountered)

                    if sub_result is None:
                        if not jedi_utils.is_scope(node):
                            return (
                                None
                                if definition and node != jedi_utils.get_parent_scope(definition)
                                else [definition]
                            )
                        else:
                            sub_result = []
                    names.extend(sub_result)
                    if self._is_global_stmt_with_name(c, name.value):
                        global_encountered = True
            elif isinstance(node, tree.Name) and node.value == name.value:
                if definition and definition != node:
                    if self._is_name_function_definition(node):
                        if isinstance(
                            jedi_utils.get_parent_scope(jedi_utils.get_parent_scope(node)),
                            tree.Class,
                        ):
                            return []
                        else:
                            return None
                    if (
                        node.is_definition()
                        and not global_encountered
                        and (
                            is_function_definition
                            or jedi_utils.get_parent_scope(node)
                            != jedi_utils.get_parent_scope(definition)
                        )
                    ):
                        return None
                    if self._is_name_function_definition(definition) and isinstance(
                        jedi_utils.get_parent_scope(jedi_utils.get_parent_scope(definition)),
                        tree.Class,
                    ):
                        return None
                names.append(node)
            return names

        if definition:
            if self._is_name_function_definition(definition):
                scope = jedi_utils.get_parent_scope(jedi_utils.get_parent_scope(definition))
            else:
                scope = jedi_utils.get_parent_scope(definition)
        else:
            scope = jedi_utils.get_parent_scope(name)

        usages = find_usages_in_node(scope)
        return usages

    def get_positions_for(self, source, line, column):
        module_node = jedi_utils.parse_source(source)
        pos = (line, column)
        stmt = self._get_statement_for_position(module_node, pos)

        name = None
        if isinstance(stmt, tree.Name):
            name = stmt
        elif isinstance(stmt, tree.BaseNode):
            name = jedi_utils.get_name_of_position(stmt, pos)

        if not name:
            return set()

        # format usage positions as tkinter text widget indices
        return set(
            (
                "%d.%d" % (usage.start_pos[0], usage.start_pos[1]),
                "%d.%d" % (usage.start_pos[0], usage.start_pos[1] + len(name.value)),
            )
            for usage in self._find_usages(name, stmt)
        )


class UsagesHighlighter(BaseNameHighlighter):
    """Script.usages looks tempting method to use for finding variable ocurrences,
    but it only returns last
    assignments to a variable, not really all usages (with Jedi 0.10).
    But it finds attribute usages quite nicely.
    
    TODO: check if this gets fixed in later versions of Jedi
    
    NB!!!!!!!!!!!!! newer jedi versions use subprocess and are too slow to run
    for each keypress"""

    def get_positions_for(self, source, line, column):
        # https://github.com/davidhalter/jedi/issues/897
        script = Script(source + ")", line=line, column=column, path="")
        usages = script.usages()

        result = {
            (
                "%d.%d" % (usage.line, usage.column),
                "%d.%d" % (usage.line, usage.column + len(usage.name)),
            )
            for usage in usages
            if usage.module_name == ""
        }

        return result


class CombinedHighlighter(VariablesHighlighter, UsagesHighlighter):
    def get_positions_for(self, source, line, column):
        usages = UsagesHighlighter.get_positions_for(self, source, line, column)
        variables = VariablesHighlighter.get_positions_for(self, source, line, column)
        return usages | variables


def update_highlighting(event):
    assert isinstance(event.widget, tk.Text)
    text = event.widget

    if not hasattr(text, "name_highlighter"):
        text.name_highlighter = VariablesHighlighter(text)
        # Alternatives:
        # NB! usages() is too slow when used on library names
        # text.name_highlighter = CombinedHighlighter(text)
        # text.name_highlighter = UsagesHighlighter(text)

    text.name_highlighter.schedule_update()


def load_plugin() -> None:
    if jedi_utils.get_version_tuple() < (0, 9):
        logging.warning("Jedi version is too old. Disabling name highlighter")
        return

    wb = get_workbench()
    wb.set_default("view.name_highlighting", False)
    wb.bind_class("CodeViewText", "<<CursorMove>>", update_highlighting, True)
    wb.bind_class("CodeViewText", "<<TextChange>>", update_highlighting, True)
    wb.bind("<<UpdateAppearance>>", update_highlighting, True)



# -*- coding: utf-8 -*-

import os
import tkinter as tk

from thonny import get_workbench, get_runner, get_shell
from thonny.base_file_browser import BaseLocalFileBrowser, BaseRemoteFileBrowser
from thonny.ui_utils import lookup_style_option
from thonny.common import normpath_with_actual_case, InlineCommand
from thonny.running import construct_cd_command
from thonny.misc_utils import running_on_windows
from tkinter import messagebox

minsize = 80


class FilesView(tk.PanedWindow):
    def __init__(self, master=None):
        tk.PanedWindow.__init__(self, master, orient="vertical", borderwidth=0)
        self.remote_added = False

        self.configure(sashwidth=lookup_style_option("Sash", "sashthickness", 4))
        self.configure(background=lookup_style_option("TPanedWindow", "background"))

        get_workbench().bind("BackendTerminated", self.on_backend_terminate, True)
        get_workbench().bind("BackendRestart", self.on_backend_restart, True)
        get_workbench().bind("WorkbenchClose", self.on_workbench_close, True)

        self.local_files = ActiveLocalFileBrowser(self)
        self.local_files.check_update_focus()
        self.add(self.local_files, minsize=minsize)

        self.remote_files = ActiveRemoteFileBrowser(self)
        self.reset_remote()

    def on_show(self):
        self.reset_remote()
        self.local_files.refresh_tree()

    def reset_remote(self, msg=None):
        runner = get_runner()
        if not runner:
            return

        proxy = runner.get_backend_proxy()
        if not proxy:
            self.hide_remote()
            return

        if proxy.has_own_filesystem():
            # remote pane is needed
            if not self.remote_added:
                self.add(self.remote_files, before=self.local_files, minsize=minsize)
                self.remote_added = True
                self.restore_split()
            self.remote_files.clear()
            self.remote_files.check_update_focus()
        else:
            # remote pane not needed
            self.hide_remote()

    def hide_remote(self):
        if self.remote_added:
            self.save_split()
            self.remove(self.remote_files)
            self.remote_added = False

    def save_split(self):
        _, y = self.sash_coord(0)
        get_workbench().set_option("view.files_split", y)

    def restore_split(self):
        split = get_workbench().get_option("view.files_split", None)
        if split is None:
            if self.winfo_height() > 5:
                split = int(self.winfo_height() * 0.66)
            else:
                split = 600

        self.sash_place(0, 0, split)

    def on_backend_restart(self, event):
        if event.get("full"):
            self.reset_remote(event)

    def on_backend_terminate(self, event):
        self.reset_remote(event)

    def on_workbench_close(self, event=None):
        if self.remote_added:
            self.save_split()

    def get_active_local_dir(self):
        return self.local_files.get_active_directory()

    def get_active_remote_dir(self):
        if self.remote_added:
            return self.remote_files.get_active_directory()
        else:
            return None

    def destroy(self):
        get_workbench().unbind("BackendTerminated", self.on_backend_terminate)
        get_workbench().unbind("BackendRestart", self.on_backend_restart)
        get_workbench().unbind("WorkbenchClose", self.on_workbench_close)
        super().destroy()


class ActiveLocalFileBrowser(BaseLocalFileBrowser):
    def __init__(self, master, show_hidden_files=False):
        super().__init__(master, show_hidden_files)
        get_workbench().bind("ToplevelResponse", self.on_toplevel_response, True)

    def create_new_file(self):
        path = super().create_new_file()
        if path and path.endswith(".py"):
            get_workbench().get_editor_notebook().show_file(path)

    def get_proposed_new_file_name(self, folder, extension):
        base = "new_file"

        if os.path.exists(os.path.join(folder, base + extension)):
            i = 2

            while True:
                name = base + "_" + str(i) + extension
                path = os.path.join(folder, name)
                if os.path.exists(path):
                    i += 1
                else:
                    return name
        else:
            return base + extension

    def request_focus_into(self, path):
        if path == "":
            if running_on_windows():
                # list of drives, can't cd
                return self.focus_into(path)
            else:
                path = "/"

        if not os.path.isdir(path):
            return

        proxy = get_runner().get_backend_proxy()
        if (
            proxy
            and proxy.uses_local_filesystem()
            and proxy.get_cwd() != path
            and get_runner().is_waiting_toplevel_command()
        ):
            get_shell().submit_magic_command(construct_cd_command(path))
        else:
            # it's OK, if it's already focused into this directory
            # focus again to refresh
            self.focus_into(path)
            get_workbench().set_local_cwd(path)

    def on_toplevel_response(self, event):
        self.check_update_focus()

    def check_update_focus(self):
        cwd = get_workbench().get_local_cwd()
        if cwd != self.current_focus and os.path.isdir(cwd):
            self.focus_into(cwd)

    def check_add_upload_command(self):
        target_dir = self.master.get_active_remote_dir()
        if target_dir is None:
            return

        proxy = get_runner().get_backend_proxy()

        if not proxy.supports_directories():
            target_dir_desc = proxy.get_node_label()
        else:
            target_dir_desc = target_dir

        def upload():
            selection = self.get_selection_info(True)
            if not selection:
                return

            if "dir" in selection["kinds"] and not proxy.supports_directories():
                messagebox.showerror(
                    "Can't upload directory",
                    "%s does not support directories.\n" % proxy.get_node_label()
                    + "You can only upload files.",
                )
            else:
                response = get_runner().send_command(
                    InlineCommand(
                        "upload",
                        allow_overwrite=False,
                        source_paths=selection["paths"],
                        target_dir=target_dir,
                        blocking=True,
                        description=_("Uploading %s to %s")
                        % (selection["description"], target_dir),
                    )
                )
                check_upload_download_response("upload", response)
                self.master.remote_files.refresh_tree()

        self.menu.add_command(label=_("Upload to %s") % target_dir_desc, command=upload)

    def add_middle_menu_items(self):
        self.check_add_upload_command()
        super().add_middle_menu_items()


class ActiveRemoteFileBrowser(BaseRemoteFileBrowser):
    def __init__(self, master, show_hidden_files=False):
        super().__init__(master, show_hidden_files)
        get_workbench().bind("ToplevelResponse", self.on_toplevel_response, True)

    def on_toplevel_response(self, event):
        if get_runner().get_backend_proxy().has_own_filesystem():
            self.check_update_focus()

    def check_update_focus(self):
        proxy = get_runner().get_backend_proxy()
        if self.current_focus != proxy.get_cwd():
            self.focus_into(proxy.get_cwd())

    def request_focus_into(self, path):
        proxy = get_runner().get_backend_proxy()
        if proxy:
            assert proxy.has_own_filesystem()

            if not get_runner().is_waiting_toplevel_command():
                messagebox.showerror(
                    "Error",
                    "Can't change or refresh directories when device is busy.\n"
                    + "Wait until current command completes and try again.",
                )
            elif not proxy.supports_directories():
                assert path == ""
                self.focus_into(path)
            elif self.current_focus == path:
                # refreshes
                self.focus_into(path)
            else:
                get_shell().submit_magic_command(["%cd", path if path != "" else "/"])

    def add_download_command(self):
        target_dir = self.master.get_active_local_dir()

        def download():
            selection = self.get_selection_info(True)
            if not selection:
                return

            response = get_runner().send_command(
                InlineCommand(
                    "download",
                    allow_overwrite=False,
                    source_paths=selection["paths"],
                    target_dir=target_dir,
                    blocking=True,
                    description=_("Downloading %s to %s") % (selection["description"], target_dir),
                )
            )
            check_upload_download_response("download", response)
            self.master.local_files.refresh_tree()

        self.menu.add_command(label=_("Download to %s") % target_dir, command=download)

    def add_middle_menu_items(self):
        self.add_download_command()
        super().add_middle_menu_items()


def check_upload_download_response(command_name, command_response):
    if command_response and command_response.get("existing_files"):
        # command was not performed because overwriting existing files need confirmation
        existing = sorted(command_response["existing_files"][:25])
        if len(command_response["existing_files"]) > 25:
            existing.append("...")

        user_response = messagebox.askokcancel(
            "Overwriting",
            "Some file(s) will be overwritten:\n\n" + "   " + "\n   ".join(existing),
            icon="info",
        )
        if not user_response:
            return

        else:
            get_runner().send_command(
                InlineCommand(
                    command_name,
                    allow_overwrite=True,
                    source_paths=command_response["source_paths"],
                    target_dir=command_response["target_dir"],
                    blocking=True,
                    description=command_response["description"],
                )
            )


def load_plugin() -> None:
    get_workbench().set_default(
        "file.last_browser_folder", normpath_with_actual_case(os.path.expanduser("~"))
    )

    get_workbench().add_view(FilesView, _("Files"), "nw")



import os

from thonny.backend import get_vm
from thonny.common import BackendEvent

_last_pos = (None, None)
_notification_is_sent = False
_LAST_POS_SETTING = "dock_user_windows.last_position"


def on_configure(event):
    global _last_pos, _notification_is_sent
    pos = event.x, event.y
    if pos != _last_pos:
        get_vm().set_option(_LAST_POS_SETTING, pos)

    if not _notification_is_sent:
        get_vm().send_message(BackendEvent("UserWindowAppeared"))
        _notification_is_sent = True


def patch_tkinter(module):
    flag_name = "has_docking_patch"
    if getattr(module, flag_name, False):
        return

    original_constructor = module.Tk.__init__

    def patched_Tk_constructor(self, *args, **kw):
        original_constructor(self, *args, **kw)

        try:
            # move window to the same place it was previously
            last_pos = get_vm().get_option(_LAST_POS_SETTING)
            if isinstance(last_pos, tuple):
                self.geometry("+%d+%d" % last_pos)

            self.wm_attributes("-topmost", 1)
            # self.overrideredirect(1)

            # I'm using bind_class because turtle._Screen later overwrites the bind handler
            self.bind_class("Tk", "<Configure>", on_configure, True)
        except Exception:
            # expected to fail when constructing Tcl for _cmd_process_gui_events
            pass

    module.Tk.__init__ = patched_Tk_constructor
    setattr(module, flag_name, True)


def patch_turtle(module):
    # Turtle needs more tweaking because it later overrides the position set in the Tk constructor
    turtle_config = getattr(module, "_CFG", None)
    if isinstance(turtle_config, dict):
        last_pos = get_vm().get_option(_LAST_POS_SETTING)
        if isinstance(last_pos, tuple):
            turtle_config["leftright"], turtle_config["topbottom"] = last_pos


def load_plugin():
    if os.environ.get("DOCK_USER_WINDOWS", "False").lower() == "true":
        vm = get_vm()
        vm.add_import_handler("tkinter", patch_tkinter)
        vm.add_import_handler("turtle", patch_turtle)



class DigitalInOut:
    ""

    def deinit():
        pass

    direction = None
    drive_mode = None
    pull = None

    def switch_to_input():
        pass

    def switch_to_output():
        pass

    value = None


class Direction:
    ""
    INPUT = None
    OUTPUT = None


class DriveMode:
    ""
    OPEN_DRAIN = None
    PUSH_PULL = None


class Pull:
    ""
    DOWN = None
    UP = None



from thonny.plugins.micropython.connection import MicroPythonConnection
from queue import Queue
import threading


class WebReplConnection(MicroPythonConnection):
    def __init__(self, url, password):
        super().__init__()
        self._url = url
        self._password = password

        # Some tricks are needed to use async library in sync program
        # use thread-safe queues to communicate with async world in another thread
        self._write_queue = Queue()
        self._connection_result = Queue()
        self._ws_thread = threading.Thread(target=self._wrap_ws_main, daemon=True)
        self._ws_thread.start()

        # Wait until connection was made
        res = self._connection_result.get()
        if res != "OK":
            raise res

    def _wrap_ws_main(self):
        import asyncio

        loop = asyncio.new_event_loop()
        loop.set_debug(True)
        loop.run_until_complete(self._ws_main())

    async def _ws_main(self):
        import asyncio

        try:
            await self._ws_connect()
        except Exception as e:
            self._connection_result.put_nowait(e)
            return

        self._connection_result.put_nowait("OK")
        await asyncio.gather(self._ws_keep_reading(), self._ws_keep_writing())

    async def _ws_connect(self):
        import asyncio
        import websockets

        self._ws = await asyncio.wait_for(websockets.connect(self._url, ping_interval=None), 3)
        print("GOT WS", self._ws)

        # read password prompt and send password
        read_chars = ""
        while read_chars != "Password: ":
            print("prelude", read_chars)
            ch = await self._ws.recv()
            print("GOT", ch)
            read_chars += ch

        print("sending password")
        await self._ws.send(self._password + "\n")
        print("sent password")

    async def _ws_keep_reading(self):
        while True:
            data = (await self._ws.recv()).encode("UTF-8")
            if len(data) == 0:
                self._error = "EOF"
                break

            self.num_bytes_received += len(data)
            self._read_queue.put(data, block=False)

    async def _ws_keep_writing(self):
        import asyncio

        while True:
            while not self._write_queue.empty():
                data = self._write_queue.get(block=False).decode("UTF-8")
                print("Wrote:", repr(data))
                await self._ws.send(data)
            # Allow reading loop to progress
            await asyncio.sleep(0.01)

    def write(self, data, block_size=32, delay=0.01):
        self._write_queue.put_nowait(data)

    async def _async_close(self):
        await self._ws.close()

    def close(self):
        """
        import asyncio
        asyncio.get_event_loop().run_until_complete(self.async_close())
        """



import os.path
import subprocess
import tkinter as tk
from shutil import which
from tkinter import messagebox, ttk

from thonny import get_workbench, running, ui_utils
from thonny.common import normpath_with_actual_case
from thonny.misc_utils import running_on_mac_os, running_on_windows
from thonny.plugins.backend_config_page import BackendDetailsConfigPage
from thonny.running import WINDOWS_EXE
from thonny.ui_utils import SubprocessDialog, askdirectory, askopenfilename, create_string_var


class CustomCPythonConfigurationPage(BackendDetailsConfigPage):
    def __init__(self, master):
        super().__init__(master)

        self._configuration_variable = create_string_var(
            get_workbench().get_option("CustomInterpreter.path")
        )

        entry_label = ttk.Label(self, text=_("Known interpreters"))
        entry_label.grid(row=0, column=0, columnspan=2, sticky=tk.W)

        self._entry = ttk.Combobox(
            self,
            exportselection=False,
            textvariable=self._configuration_variable,
            values=self._get_interpreters(),
        )

        self._entry.grid(row=1, column=0, columnspan=2, sticky=tk.NSEW)
        self._entry.state(["!disabled", "readonly"])

        another_label = ttk.Label(self, text=_("Your interpreter isn't in the list?"))
        another_label.grid(row=2, column=0, columnspan=2, sticky=tk.W, pady=(10, 0))

        ttk.Style().configure("Centered.TButton", justify="center")
        self._select_button = ttk.Button(
            self,
            style="Centered.TButton",
            text=_("Locate another")
            + " "
            + ("python.exe ..." if running_on_windows() else _("python executable") + " ...")
            + "\n"
            + _("NB! Thonny only supports Python 3.5 and later"),
            command=self._select_executable,
        )

        self._select_button.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW)

        self._venv_button = ttk.Button(
            self,
            style="Centered.TButton",
            text=_("Create new virtual environment")
            + " ...\n"
            + "("
            + _("Select existing or create a new empty directory")
            + ")",
            command=self._create_venv,
        )

        self._venv_button.grid(row=4, column=0, columnspan=2, sticky=tk.NSEW, pady=(5, 0))

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

    def _select_executable(self):
        # TODO: get dir of current interpreter
        options = {"master": self}
        if running_on_windows():
            options["filetypes"] = [
                (_("Python interpreters"), "python.exe"),
                (_("all files"), ".*"),
            ]

        filename = askopenfilename(**options)

        if filename:
            self._configuration_variable.set(filename)

    def _create_venv(self):
        path = None
        while True:
            path = askdirectory(
                master=self,
                initialdir=path,
                title=_("Select empty directory for new virtual environment"),
            )
            if not path:
                return

            if os.listdir(path):
                messagebox.showerror(
                    _("Bad directory"),
                    _("Selected directory is not empty.\nSelect another or cancel."),
                )
            else:
                break
        assert os.path.isdir(path)
        path = normpath_with_actual_case(path)

        proc = subprocess.Popen(
            [running.get_frontend_python(), "-m", "venv", path],
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        dlg = SubprocessDialog(self, proc, _("Creating virtual environment"))
        ui_utils.show_dialog(dlg)

        if running_on_windows():
            exe_path = normpath_with_actual_case(os.path.join(path, "Scripts", "python.exe"))
        else:
            exe_path = os.path.join(path, "bin", "python3")

        if os.path.exists(exe_path):
            self._configuration_variable.set(exe_path)

    def _get_interpreters(self):
        result = set()

        if running_on_windows():
            # registry
            result.update(self._get_interpreters_from_windows_registry())

            for minor in [5, 6, 7, 8]:
                for dir_ in [
                    "C:\\Python3%d" % minor,
                    "C:\\Python3%d-32" % minor,
                    "C:\\Python3%d-64" % minor,
                    "C:\\Program Files\\Python 3.%d" % minor,
                    "C:\\Program Files\\Python 3.%d-64" % minor,
                    "C:\\Program Files (x86)\\Python 3.%d" % minor,
                    "C:\\Program Files (x86)\\Python 3.%d-32" % minor,
                ]:
                    path = os.path.join(dir_, WINDOWS_EXE)
                    if os.path.exists(path):
                        result.add(normpath_with_actual_case(path))

            # other locations
            for dir_ in ["C:\\Anaconda3", os.path.expanduser("~/Anaconda3")]:
                path = os.path.join(dir_, WINDOWS_EXE)
                if os.path.exists(path):
                    result.add(normpath_with_actual_case(path))

        else:
            # Common unix locations
            dirs = ["/bin", "/usr/bin", "/usr/local/bin", os.path.expanduser("~/.local/bin")]
            for dir_ in dirs:
                # if the dir_ is just a link to another dir_, skip it
                # (not to show items twice)
                # for example on Fedora /bin -> usr/bin
                if not os.path.exists(dir_):
                    continue

                apath = normpath_with_actual_case(dir_)
                if apath != dir_ and apath in dirs:
                    continue
                for name in ["python3", "python3.5", "python3.6", "python3.7", "python3.8"]:
                    path = os.path.join(dir_, name)
                    if os.path.exists(path):
                        result.add(path)

        if running_on_mac_os():
            for version in ["3.5", "3.6", "3.7", "3.8"]:
                dir_ = os.path.join("/Library/Frameworks/Python.framework/Versions", version, "bin")
                path = os.path.join(dir_, "python3")

                if os.path.exists(path):
                    result.add(path)

        for command in ["python3", "python3.5", "python3.5", "python3.6", "python3.7", "python3.8"]:
            path = which(command)
            if path is not None and os.path.isabs(path):
                result.add(path)

        for path in get_workbench().get_option("CustomInterpreter.used_paths"):
            if os.path.exists(path):
                result.add(normpath_with_actual_case(path))

        return sorted(result)

    def _get_interpreters_from_windows_registry(self):
        # https://github.com/python/cpython/blob/master/Tools/msi/README.txt
        import winreg

        result = set()
        for key in [winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER]:
            for version in [
                "3.5",
                "3.5-32",
                "3.5-64",
                "3.6",
                "3.6-32",
                "3.6-64",
                "3.7",
                "3.7-32",
                "3.7-64",
                "3.8",
                "3.8-32",
                "3.8-64",
            ]:
                try:
                    for subkey in [
                        "SOFTWARE\\Python\\PythonCore\\" + version + "\\InstallPath",
                        "SOFTWARE\\Python\\PythonCore\\Wow6432Node\\" + version + "\\InstallPath",
                    ]:
                        dir_ = winreg.QueryValue(key, subkey)
                        if dir_:
                            path = os.path.join(dir_, WINDOWS_EXE)
                            if os.path.exists(path):
                                result.add(path)
                except Exception:
                    pass

        return result

    def should_restart(self):
        return self._configuration_variable.modified  # pylint: disable=no-member

    def apply(self):
        if not self.should_restart():
            return

        path = self._configuration_variable.get()
        if os.path.isfile(path):
            get_workbench().set_option("CustomInterpreter.path", path)



class PWMOut:
    ""


class PulseIn:
    ""

    def clear():
        pass

    def deinit():
        pass

    maxlen = None

    def pause():
        pass

    paused = None

    def popleft():
        pass

    def resume():
        pass


class PulseOut:
    ""

    def deinit():
        pass

    def send():
        pass



import io
import token
import tokenize

from thonny import get_workbench
from thonny.codeview import CodeViewText
from thonny.shell import ShellText
import time

_OPENERS = {")": "(", "]": "[", "}": "{"}

TOKTYPES = {token.LPAR, token.RPAR, token.LBRACE, token.RBRACE, token.LSQB, token.RSQB}

BLANK_LINE_REGEX_STR = r"\n\s*\r?\n"
BLANK_LINE_REGEX_STR = r"^\s*$"

BLOCK_START_REGEX_STR = r"^\s*(class|def|while|elif|with|try|except|finally) "


class ParenMatcher:
    def __init__(self, text):
        self.text = text
        self._update_scheduling_id = None
        self._delayed_scheduling_id = None
        self._tokens_cache = {}

    def schedule_update(self, delay=None):
        if self._update_scheduling_id is not None:
            self.text.after_cancel(self._update_scheduling_id)

        if delay is not None:
            self._update_scheduling_id = self.text.after(delay, self.perform_update)
        else:
            self._update_scheduling_id = self.text.after_idle(self.perform_update)

    def perform_update(self):
        try:
            self.update_highlighting()
        finally:
            self._update_scheduled = False

    def invalidate_token_cache(self):
        self._tokens_cache = {}

    def update_highlighting(self):
        clear_highlighting(self.text)

        if get_workbench().get_option("view.paren_highlighting"):
            self._update_highlighting_for_active_range()

    def _update_highlighting_for_active_range(self):
        start_index = "1.0"
        end_index = self.text.index("end")

        # Try to reduce search range for better performance.
        index = self._find_block_start("@0,0 linestart", True)
        if index:
            start_index = index

        lower_right = "@%d,%d" % (self.text.winfo_width(), self.text.winfo_height())
        index = self._find_block_start(lower_right + " lineend", False)
        if index:
            end_index = index

        self._highlight(start_index, end_index)

    def _find_block_start(self, start_position, backwards):
        while True:
            index = self.text.search(
                BLOCK_START_REGEX_STR,
                start_position,
                regexp=True,
                backwards=backwards,
                stopindex="1.0" if backwards else "end",
            )
            if not index:
                break

            tags = self.text.tag_names(index)
            if "string3" in tags or "open_string3" in tags:
                # not a block start
                if backwards:
                    start_position = index
                else:
                    start_position = index + " +1c"
            else:
                break

        return index

    def _highlight(self, start_index, end_index):
        stack = []

        cursor_row, cursor_col = map(int, self.text.index("insert").split("."))

        for t in self._get_paren_tokens(start_index, end_index):
            if t.string in "([{":
                stack.append(t)
            elif not stack:
                # stack is empty, ie. found a closer without opener
                close_index = "%d.%d" % (t.start[0], t.end[1])
                self.text.tag_add("unclosed_expression", start_index, close_index)
                break
            elif stack[-1].string != _OPENERS[t.string]:
                # incorrect closure
                opener = stack[-1]
                open_index = "%d.%d" % opener.start
                self.text.tag_add("unclosed_expression", open_index, end_index)
                break
            else:
                # found a pair
                opener = stack[-1]
                closer = t

                # if cursor is right after opener or closer then highlight both
                if (
                    cursor_row == opener.start[0]
                    and cursor_col == opener.end[1]
                    or cursor_row == closer.start[0]
                    and cursor_col == closer.end[1]
                ):
                    self.text.tag_add("surrounding_parens", "%d.%d" % closer.start)
                    self.text.tag_add("surrounding_parens", "%d.%d" % opener.start)

                stack.pop()

        if stack:
            # something was left without closure
            opener = stack[-1]
            open_index = "%d.%d" % opener.start
            self.text.tag_add("unclosed_expression", open_index, end_index)

    def _get_paren_tokens(self, start_index, end_index):
        if (start_index, end_index) in self._tokens_cache:
            return self._tokens_cache[(start_index, end_index)]

        start_row, start_col = map(int, start_index.split("."))
        source = self.text.get(start_index, end_index)
        # prepend source with empty lines and spaces to make
        # token rows and columns match with widget indices
        source = ("\n" * (start_row - 1)) + (" " * start_col) + source

        result = []
        try:
            tokens = tokenize.tokenize(io.BytesIO(source.encode("utf-8")).readline)
            for token in tokens:
                # if token.string != "" and token.string in "()[]{}":
                if token.exact_type in TOKTYPES:
                    result.append(token)
        except Exception:
            # happens eg when parens are unbalanced or there is indentation error or ...
            pass

        if start_index == "1.0" and end_index == "end":
            self._tokens_cache[(start_index, end_index)] = result

        return result


class ShellParenMatcher(ParenMatcher):
    def _update_highlighting_for_active_range(self):

        # TODO: check that cursor is in this range
        index_parts = self.text.tag_prevrange("command", "end")

        if index_parts:
            start_index, end_index = index_parts
            self._highlight(start_index, end_index)


def _update_highlighting(event, text_changed, need_update, delay=None):
    text = event.widget
    if not hasattr(text, "paren_matcher"):
        if isinstance(text, CodeViewText):
            text.paren_matcher = ParenMatcher(text)
        elif isinstance(text, ShellText):
            text.paren_matcher = ShellParenMatcher(text)
        else:
            return

    if text_changed:
        text.paren_matcher.invalidate_token_cache()

    if need_update:
        text.paren_matcher.schedule_update(delay)


def update_highlighting_full(event):
    _update_highlighting(event, True, True)


def clear_highlighting(text):
    text.tag_remove("surrounding_parens", "0.1", "end")
    text.tag_remove("unclosed_expression", "0.1", "end")


_last_move_time = 0


def update_highlighting_move(event):
    global _last_move_time
    # needs delay because selecting with mouse causes many events
    # and I don't know how to distinguish selection from other moves
    t = time.time()
    if t - _last_move_time > 0.1:
        delay = None
    else:
        delay = 300
    _last_move_time = t
    _update_highlighting(event, False, True, delay=delay)


def update_highlighting_edit_cw(event):
    if isinstance(event.text_widget, CodeViewText):
        event.widget = event.text_widget
        trivial = event.get("trivial_for_parens", False)
        _update_highlighting(event, True, not trivial)
        if trivial:
            event.text_widget.tag_remove("surrounding_parens", "0.1", "end")


def load_plugin() -> None:
    wb = get_workbench()

    wb.set_default("view.paren_highlighting", True)
    wb.bind("TextInsert", update_highlighting_edit_cw, True)
    wb.bind("TextDelete", update_highlighting_edit_cw, True)
    wb.bind_class("CodeViewText", "<<VerticalScroll>>", update_highlighting_move, True)
    wb.bind_class("CodeViewText", "<<CursorMove>>", update_highlighting_move, True)
    wb.bind_class("ShellText", "<<TextChange>>", update_highlighting_full, True)
    wb.bind_class("ShellText", "<<CursorMove>>", update_highlighting_full, True)
    wb.bind("<<UpdateAppearance>>", update_highlighting_full, True)
